#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Generator, Optional, Union
from pathlib import Path
from zipfile import ZipFile

import mimetypes
import zipfile
import tarfile
import gzip
import shutil
import logging


class InvalidFileExtension(Exception):

    """
        Exception raised if file extension is unkown or invalid
    """

    pass


def guessMimeType(filePath: Union[Path, str]) -> str:
    """
        Tries to guess mime type of the file

        Parameters
        ----------
        filePath : Union[Path, str]
            file whose mime type will be guessed

        Returns
        -------
        str -> guessed mime type, or "application/octet-stream" if
        it was not possible to guess
    """

    mimeTypesResult = mimetypes.guess_type(filePath)

    mimeType = mimeTypesResult[0]
    if mimeType is None:
        return "application/octet-stream"

    return mimeType


def isGzip(path: Path) -> bool:
    """
        Checks if the file is compressed with gz

        Might not be 100% reliable, it checks the first 2 bytes
        of the file for the gz-compressed file header (0x1F and 0x8B)
        and checks for .gz file extension

        Parameters
        ----------
        path : Path
            the file to be checked

        Returns
        -------
        bool -> True if file is gz compressed, False otherwise
    """

    # .gz compressed files always start with 2 bytes: 0x1F and 0x8B
    # Testing for this is not 100% reliable, it is highly unlikely
    # that "ordinary text files" start with those two bytes—in UTF-8 it's not even legal.
    # That's why we check for extension and do the byte checking
    # Ref: https://stackoverflow.com/a/3703300/7585106

    if not path.is_file():
        return False

    with open(path, 'rb') as file:
        return file.read(2) == b'\x1f\x8b' and path.name.endswith(".gz")


def isArchive(path: Path) -> bool:
    """
        Checks if the file is an archive

        Parameters
        ----------
        path : Path
            file to be checked

        Returns
        -------
        bool -> True if it is an archive, False otherwise
    """

    return zipfile.is_zipfile(path) or tarfile.is_tarfile(path)


def gzipDecompress(source: Path, destination: Path) -> None:
    """
        Decompresses a gz-compressed file

        Parameters
        ----------
        source : Path
            file to be decompressed
        destination : Path
            location to which the decompressed file will be stored

        Raises
        ------
        ValueError -> if the file is not a gz-compressed file
    """

    if not isGzip(source):
        raise ValueError(">> [Coretex] Not a .gz file")

    with gzip.open(source, "r") as gzipFile, open(destination, "wb") as destinationFile:
        shutil.copyfileobj(gzipFile, destinationFile)


def archive(source: Path, destination: Path) -> None:
    """
        Archives and compresses the provided file or directory
        using ZipFile module

        Parameters
        ----------
        source : Path
            file to be archived and compressed
        destination : Path
            location to which the zip file will be stored
    """

    with ZipFile(destination, "w", zipfile.ZIP_DEFLATED) as destinationFile:
        if source.is_file():
            destinationFile.write(source, source.name)
        else:
            for path in source.rglob("*"):
                if not path.is_file():
                    continue

                destinationFile.write(path, path.relative_to(source))


def walk(path: Path) -> Generator[Path, None, None]:
    """
        os.walk implementation for pathlib.Path

        Parameters
        ----------
        path : Path
            starting point of the walk function, must be a directory

        Returns
        -------
        Generator[Path, None, None] -> generator which contains all
        subdirectories and subfiles
    """

    for p in path.iterdir():
        yield p.resolve()

        if p.is_dir():
            yield from walk(p)


def recursiveUnzip(entryPoint: Path, destination: Optional[Path] = None, remove: bool = False) -> None:
    """
        Recursively unarchives the file

        Parameters
        ----------
        entryPoint : Path
            initial archive
        destination : Optional[Path]
            destination of unarchived files
        remove : bool
            delete archive after unarchive is done

        Raises
        ------
        ValueError -> if the path is not an archive
    """

    logging.getLogger("coretexpylib").debug(f">> [Coretex] recursiveUnzip: source = {str(entryPoint)}, destination = {str(destination)}")

    if destination is None:
        destination = entryPoint.parent / entryPoint.stem

    # Decompress with gzip if is gzip
    if isGzip(entryPoint):
        gzipDecompress(entryPoint, destination)

        if remove:
            entryPoint.unlink()

        if not isArchive(destination):
            return

        # gzip nameing convention is .original_file_ext.gz, so by calling .stem we remove .gz
        # for destination
        recursiveUnzip(destination, destination.parent / destination.stem, remove = True)
        return

    if not isArchive(entryPoint):
        raise ValueError(">> [Coretex] Not an archive")

    if zipfile.is_zipfile(entryPoint):
        with ZipFile(entryPoint, "r") as zipFile:
            zipFile.extractall(destination)

    if tarfile.is_tarfile(entryPoint):
        with tarfile.open(entryPoint, "r") as tarFile:
            tarFile.extractall(destination)

    if remove:
        entryPoint.unlink()

    for path in walk(destination):
        if isArchive(path) or isGzip(path):
            recursiveUnzip(path, remove = True)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Callable
from threading import Thread, Event, RLock

import time
import logging

from ...logging import Log


MAX_WAIT_TIME_BEFORE_UPDATE = 5
UploadFunction = Callable[[List[Log]], bool]


class LoggerUploadWorker(Thread):

    """
        Not intended for outside use

        A worker thread which is constantly running and
        uploading logs to Coretex backend every 5 seconds

        If the upload request fails the wait time is doubled
    """

    def __init__(self, uploadFunction: UploadFunction) -> None:
        super().__init__()

        self.setDaemon(True)
        self.setName("LoggerUploadWorker")

        self.__uploadFunction = uploadFunction
        self.__stopped = Event()
        self.__lock = RLock()
        self.__waitTime = MAX_WAIT_TIME_BEFORE_UPDATE
        self.__pendingLogs: List[Log] = []

    @property
    def isStopped(self) -> bool:
        return self.__stopped.is_set()

    def stop(self) -> None:
        with self.__lock:
            self.__stopped.set()

    def add(self, log: Log) -> None:
        with self.__lock:
            if self.isStopped:
                return

            self.__pendingLogs.append(log)

    def uploadLogs(self) -> bool:
        with self.__lock:
            if self.isStopped:
                return False

            if len(self.__pendingLogs) == 0:
                return True

            # Uploads logs to Coretex using the provided function
            result = self.__uploadFunction(self.__pendingLogs)

            # Only clear logs if they were successfully uploaded to coretex
            if result:
                self.__pendingLogs.clear()

            return result

    def run(self) -> None:
        while not self.isStopped:
            time.sleep(self.__waitTime)

            try:
                success = self.uploadLogs()
            except BaseException as exception:
                logging.getLogger("coretexpylib").debug(">> [Coretex] Failed to upload logs", exc_info = exception)
                success = False

            if success:
                # If upload of logs was success reset wait time
                self.__waitTime = MAX_WAIT_TIME_BEFORE_UPDATE
            else:
                # If upload of logs failed, double the wait time
                self.__waitTime *= 2



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from ..entities import TaskRun


class _CurrentTaskRunContainer:

    taskRun: Optional[TaskRun] = None


def setCurrentTaskRun(taskRun: Optional[TaskRun]) -> None:
    _CurrentTaskRunContainer.taskRun = taskRun


def currentTaskRun() -> TaskRun:
    """
        Returns
        -------
        TaskRun -> Currently executing Task
    """

    taskRun = _CurrentTaskRunContainer.taskRun
    if taskRun is None:
        raise ValueError("TaskRun is not currently executing")

    return taskRun



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, List, Any
from typing_extensions import Self, override
from pathlib import Path

from .base import BaseImageDataset
from ..network_dataset import NetworkDataset
from ...sample import ImageSample
from ...annotation import ImageDatasetClass, ImageDatasetClasses
from ....codable import KeyDescriptor, Codable
from ....networking import networkManager, FileData, NetworkRequestError


class ClassDistribution(Codable):

    name: str
    color: str
    count: int


class ImageDataset(BaseImageDataset[ImageSample], NetworkDataset[ImageSample]):  # type: ignore

    """
        Represents the Image Dataset class \n
        Includes functionality for working with Image Datasets
        that are uploaded to Coretex.ai
    """

    classDistribution: List[ClassDistribution]

    def __init__(self) -> None:
        super().__init__(ImageSample)

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()

        descriptors["samples"] = KeyDescriptor("sessions", ImageSample, list)
        descriptors["classes"] = KeyDescriptor("classes", ImageDatasetClass, ImageDatasetClasses)
        descriptors["classDistribution"] = KeyDescriptor("class_distribution", ClassDistribution, list)

        return descriptors

    @classmethod
    def fetchById(cls, objectId: int, **kwargs: Any) -> Self:
        obj = super().fetchById(objectId, **kwargs)

        response = networkManager.get(f"annotation-class?dataset_id={obj.id}")

        if not response.hasFailed():
            obj.classes = cls._decodeValue("classes", response.getJson(list))
            obj._writeClassesToFile()

        return obj

    def saveClasses(self, classes: ImageDatasetClasses) -> bool:
        parameters = {
            "dataset_id": self.id,
            "classes": [clazz.encode() for clazz in classes]
        }

        response = networkManager.post("annotation-class", parameters)
        if not response.hasFailed():
            return super().saveClasses(classes)

        return not response.hasFailed()

    @override
    def _uploadSample(self, samplePath: Path, sampleName: str, **metadata: Any) -> ImageSample:
        params = {
            "dataset_id": self.id
        }

        files = [
            FileData.createFromPath("file", samplePath)
        ]

        response = networkManager.formData("session/import", params, files)
        if response.hasFailed():
            raise NetworkRequestError(response, f"Failed to create image Sample from \"{samplePath}\"")

        return self._sampleType.decode(response.getJson(dict))



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Dict
from typing_extensions import Self

from .base import BaseObject
from ...codable import KeyDescriptor


class Task(BaseObject):

    """
        Represents the task entity from Coretex.ai\n
        Contains properties that describe the task
    """

    isDefault: bool
    taskId: int

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()
        descriptors["taskId"] = KeyDescriptor("parentId")

        return descriptors

    @classmethod
    def createTask(cls, name: str, projectId: int, description: Optional[str]=None) -> Optional[Self]:
        """
            Creates a new task with the provided name and description
            Task is added to the project with provided project id

            Parameters
            ----------
            name : str
                task name
            projectId : int
                project id the task belongs to
            description : Optional[str]
                task description

            Returns
            -------
            Optional[Self] -> The created task object

            Example
            -------
            >>> from coretex import Task
            \b
            >>> dummyTask = Task.createTask(
                    name = "dummyTask",
                    projectId = 23,
                    description = "This is dummy task"
                )
            >>> if dummyTask is None:
                    print("Failed to create task")
        """

        return cls.create(
            name = name,
            parent_id = projectId,
            description = description
        )



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

# from typing import List, Dict
# from typing_extensions import Self
# from pathlib import Path
# from zipfile import ZipFile

# import json

# from .token import Token
# from ..codable import Codable, KeyDescriptor


# class Transcription(Codable):

#     """
#         Represents a result of a single transcription

#         Properties:
#         text: str -> transcribed text
#         tokens: List[Token] -> tokens of transcribed text
#     """

#     text: str
#     tokens: List[Token]

#     @classmethod
#     def create(cls, text: str, tokens: List[Token]) -> Self:
#         obj = cls()

#         obj.text = text
#         obj.tokens = tokens

#         return obj

#     @classmethod
#     def load(cls, filePath: Path) -> Self:
        # """
        #     Loads saved transcription from a json file

        #     Parameters
        #     ----------
        #     filePath : Path
        #         path to json file

        #     Returns
        #     -------
        #     Self -> loaded transcription
        # """

#         with filePath.open("r") as file:
#             return cls.decode(json.load(file))

#     @classmethod
#     def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
#         descriptors = super()._keyDescriptors()
#         descriptors["tokens"] = KeyDescriptor("tokens", Token, list)

#         return descriptors

#     def save(self, path: Path) -> Path:
        # """
        #     Saves transcription to a zipped json file

        #     Parameters
        #     ----------
        #     path : Path
        #         path to json file

        #     Returns
        #     -------
        #     Path -> path to zip file
        # """

#         with path.open("w") as file:
#             json.dump(self.encode(), file, indent = 4)

#         zipPath = path.parent / f"{path.stem}.zip"

#         with ZipFile(zipPath, "w") as zipFile:
#             zipFile.write(str(path), path.name)

#         return zipPath



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Tuple
from typing_extensions import Self
from pathlib import Path
from zipfile import ZipFile

import copy
import json

from PIL import Image
from PIL.Image import Image as PILImage
from numpy import ndarray

import numpy as np

from .base import BaseImageDataset
from ...sample import ImageSample, AnnotatedImageSampleData
from ...annotation import CoretexSegmentationInstance, CoretexImageAnnotation, BBox
from .... import folder_manager


ANNOTATION_NAME = "annotations.json"


class AugmentedImageSample(ImageSample):

    @property
    def path(self) -> Path:
        """
            Returns
            -------
            Path -> path to new augmented samples directory
        """

        return folder_manager.temp / "temp-augmented-ds" / str(self.id)

    @classmethod
    def createFromSample(cls, sample: ImageSample) -> Self:
        """
            Creates exact copy of sample from provided sample

            Parameters
            ----------
            sample : ImageSample
                sample object

            Returns
            -------
            Self -> sample object
        """

        obj = cls()

        for key, value in sample.__dict__.items():
            obj.__dict__[key] = copy.deepcopy(value)

        return obj


def isOverlapping(
    x: int,
    y: int,
    image: PILImage,
    locations: List[Tuple[int, int, int, int]]
) -> bool:

    for loc in locations:
        if (x < loc[0] + loc[2] and x + image.width > loc[0] and
            y < loc[1] + loc[3] and y + image.height > loc[1]):

            return True

    return False


def generateSegmentedImage(image: np.ndarray, segmentationMask: np.ndarray) -> Image:
    rgbaImage = Image.fromarray(image).convert("RGBA")

    segmentedImage = np.asarray(rgbaImage) * segmentationMask
    segmentedImage = Image.fromarray(segmentedImage)

    alpha = segmentedImage.getchannel("A")
    bbox = alpha.getbbox()
    croppedImage = segmentedImage.crop(bbox)

    return croppedImage


def composeImage(
    segmentedImages: List[PILImage],
    backgroundImage: np.ndarray,
    angle: int,
    scale: float
) -> Tuple[PILImage, List[Tuple[int, int]]]:

    centroids: List[Tuple[int, int]] = []
    locations: List[Tuple[int, int, int, int]] = []

    background = Image.fromarray(backgroundImage)

    for segmentedImage in segmentedImages:
        image = segmentedImage

        rotatedImage = image.rotate(angle, expand = True)
        resizedImage = rotatedImage.resize((int(rotatedImage.width * scale), int(rotatedImage.height * scale)))

        # Calculate the maximum x and y coordinates for the top left corner of the image
        maxX = background.width - resizedImage.width
        maxY = background.height - resizedImage.height

        while True:
            # Generate a random location within the bounds of the background image
            x = np.random.randint(0, maxX)
            y = np.random.randint(0, maxY)

            # Check if the image overlaps with any previously pasted images
            if not isOverlapping(x, y, resizedImage, locations):
                break

        background.paste(resizedImage, (x, y), resizedImage)

        centerX = x + resizedImage.width // 2
        centerY = y + resizedImage.height // 2

        centroids.append((centerX, centerY))

        # Add the location to the list
        locations.append((x, y, resizedImage.width, resizedImage.height))

    return background, centroids


def processInstance(
    sample: ImageSample,
    backgroundSampleData: AnnotatedImageSampleData,
    angle: int,
    scale: float
) -> Tuple[PILImage, List[CoretexSegmentationInstance]]:

    segmentedImages: List[PILImage] = []
    augmentedInstances: List[CoretexSegmentationInstance]= []

    sampleData = sample.load()
    if sampleData.annotation is None:
        raise RuntimeError(f"CTX sample dataset sample id: {sample.id} image doesn't exist!")

    for instance in sampleData.annotation.instances:
        foregroundMask = instance.extractBinaryMask(sampleData.image.shape[1], sampleData.image.shape[0])
        segmentedImage = generateSegmentedImage(sampleData.image, foregroundMask)

        segmentedImages.append(segmentedImage)

    composedImage, centroids = composeImage(segmentedImages, backgroundSampleData.image, angle, scale)

    for i, instance in enumerate(sampleData.annotation.instances):
        segmentationsFlatten = [sample for sublist in instance.segmentations for sample in sublist]
        augmentedInstance = CoretexSegmentationInstance.create(instance.classId, BBox.fromPoly(segmentationsFlatten), instance.segmentations)

        augmentedInstance.rotateSegmentations(angle)
        augmentedInstance.centerSegmentations(centroids[i])

        augmentedInstances.append(augmentedInstance)

    return composedImage, augmentedInstances


def processSample(
    sample: ImageSample,
    backgroundSample: ImageSample,
    angle: int,
    scale: float
) -> Tuple[ndarray, CoretexImageAnnotation]:

    backgroundSampleData = backgroundSample.load()

    composedImage, augmentedInstances = processInstance(sample, backgroundSampleData, angle, scale)
    annotation = CoretexImageAnnotation.create(sample.name, composedImage.width, composedImage.height, augmentedInstances)

    return composedImage, annotation


def storeFiles(
    tempPath: Path,
    augmentedSample: ImageSample,
    augmentedImage: PILImage,
    annotation: CoretexImageAnnotation
) -> None:

    imagePath = tempPath / f"{augmentedSample.name}.jpeg"
    annotationPath = tempPath / ANNOTATION_NAME
    zipPath = tempPath / f"{augmentedSample.id}"

    augmentedImage.save(imagePath)

    with open(annotationPath, 'w') as annotationfile:
        jsonObject = json.dumps(annotation.encode())
        annotationfile.write(jsonObject)

    with ZipFile(zipPath.with_suffix(".zip"), mode = "w") as archive:
        archive.write(imagePath, f"{augmentedSample.name}.jpeg")
        archive.write(annotationPath, ANNOTATION_NAME)

    imagePath.unlink(missing_ok = True)
    annotationPath.unlink(missing_ok = True)


def augmentDataset(
    normalDataset: BaseImageDataset,
    backgroundDataset: BaseImageDataset,
    angle: int = 0,
    scale: float = 1.0
) -> None:
    """
        Modifies normalDataset by adding new augmented samples to it

        Parameters
        ----------
        normalDataset : BaseImageDataset
            BaseImageDataset object
        backgroundDataset : BaseImageDataset
            BaseImageDataset object
        angle : int
            angle of rotation in degrees
        scale : float
            scaling factor
    """

    tempPath = folder_manager.createTempFolder("temp-augmented-ds")
    augmentedSamples: List[AugmentedImageSample] = []

    for i, background in enumerate(backgroundDataset.samples):
        background.unzip()

        for j, sample in enumerate(normalDataset.samples):
            sample.unzip()
            augmentedImage, annotations = processSample(sample, background, angle, scale)

            augmentedSample = AugmentedImageSample.createFromSample(sample)
            augmentedSample.id = int(f"{i}{j}{augmentedSample.id}")

            storeFiles(tempPath, augmentedSample, augmentedImage, annotations)

            augmentedSamples.append(augmentedSample)

    for sample in augmentedSamples:
        normalDataset.samples.append(sample)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

# from typing import Optional
# from pathlib import Path

# from ..coretex import CustomSample
# from ..utils import guessMimeType, InvalidFileExtension


# def getTxtFilePath(sample: CustomSample) -> Optional[Path]:
    # """
    #     Looks for text file in Coretex sample
    #     If extension is .txt file is considered textual

    #     Parameters
    #     ----------
    #     sample : CustomSample
    #         sample to be searched

    #     Returns
    #     -------
    #     Optional[Path] -> first occurence of .txt file if there is any, None otherwise
    # """

#     folderContent = list(sample.load().folderContent)
#     for path in folderContent:
#         if path.suffix == ".txt":
#             return path

#     return None


# def isTxtSample(sample: CustomSample) -> bool:
    # """
    #     Checks whether sample is text sample or not

    #     Parameters
    #     ----------
    #     sample : CustomSample
    #         sample to be checked

    #     Returns
    #     -------
    #     bool -> True if sample is text sample, False otherwise
    # """

#     return getTxtFilePath(sample) is not None


# def getAudioFilePath(sample: CustomSample) -> Optional[Path]:
    # """
    #     Looks for any kind of audio file in Coretex sample
    #     Guesses mime type of the file and looks for any kind of audio mime type

    #     Parameters
    #     ----------
    #     sample : CustomSample
    #         sample to be searched

    #     Returns
    #     -------
    #     Optional[Path] -> first occurence of audio file if there is any, None otherwise
    # """

#     folderContent = list(sample.load().folderContent)
#     for path in folderContent:
#         try:
#             mimeType = guessMimeType(str(path))
#             if "audio" in mimeType:
#                 return path
#         except InvalidFileExtension:
#             continue

#     return None


# def isAudioSample(sample: CustomSample) -> bool:
    # """
    #     Checks whether sample is audio sample or not

    #     Parameters
    #     ----------
    #     sample : CustomSample
    #         sample to be checked

    #     Returns
    #     -------
    #     bool -> True if sample is audio sample, False otherwise
    # """

#     return getAudioFilePath(sample) is not None



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from cryptography.hazmat.primitives import hashes


class ByteBuffer:

    def __init__(self, data: bytes) -> None:
        self.data = data
        self.position = 0

    def update(self, data: bytes) -> None:
        """
            Sets the internal state of the buffer
            to the provided bytes array

            Parameters
            ----------
            data : bytes
                new internal state of the buffer
        """

        self.data = data
        self.position = 0

    def get(self) -> int:
        """
            Retrieves the byte at the current position, and
            increments internal state

            Returns
            -------
            int -> byte at the current position

            Raises
            ------
            OverflowError -> if the bytes have been exhausted
        """

        if self.position >= len(self.data):
            raise OverflowError

        byte = self.data[self.position]
        self.position += 1

        return byte


class Random:

    """
        Random generator based on recursive hashing the provided seed
        using the provided hashing algorithm. Entropy is equal to the
        provided seed.
    """

    def __init__(self, seed: bytes, algorithm: hashes.HashAlgorithm) -> None:
        self.algorithm = algorithm

        numberGenerator = hashes.Hash(algorithm)
        numberGenerator.update(seed)

        self._currentState = numberGenerator.finalize()
        self._buffer = ByteBuffer(self._currentState)

    def getRandomByte(self) -> int:
        """
            Returns
            -------
            int -> randomly generated byte value
        """

        try:
            return self._buffer.get()
        except OverflowError:
            # Raised when all bytes from current state have been exhausted
            # so we need to generate new bytes
            numberGenerator = hashes.Hash(self.algorithm)
            numberGenerator.update(self._currentState)

            self._currentState = numberGenerator.finalize()
            self._buffer.update(self._currentState)

            return self.getRandomByte()

    def getRandomBytes(self, count: int) -> bytes:
        """
            Parameters
            ----------
            count : int
                # of how many bytes should be generated

            Returns
            -------
            bytes -> randomly generated bytes array
        """

        values = bytearray(count)

        for i in range(count):
            values[i] = self.getRandomByte()

        return values



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, TypeVar, Generic, List, Dict, Any, Type, Union
from typing_extensions import Self
from datetime import datetime
from pathlib import Path
from abc import ABC, abstractmethod
from contextlib import ExitStack

import hashlib
import base64
import logging

from .dataset import Dataset
from .state import DatasetState
from ..sample import NetworkSample
from ... import folder_manager
from ...codable import KeyDescriptor
from ...networking import NetworkObject, \
    fileChunkUpload, networkManager, NetworkRequestError
from ...threading import MultithreadedDataProcessor
from ...cryptography import aes, getProjectKey
from ...utils.file import isArchive, archive


SampleType = TypeVar("SampleType", bound = "NetworkSample")
MAX_DATASET_NAME_LENGTH = 50


def _hashDependencies(dependencies: List[str]) -> str:
    hash = hashlib.md5()
    hash.update("".join(sorted(dependencies)).encode())

    return base64.b64encode(hash.digest()).decode("ascii").replace("+", "0")


def _chunkSampleImport(sampleType: Type[SampleType], sampleName: str, samplePath: Path, datasetId: int) -> SampleType:
    parameters = {
        "name": sampleName,
        "dataset_id": datasetId,
        "file_id": fileChunkUpload(samplePath)
    }

    response = networkManager.formData("session/import", parameters, timeout = (5, 300))
    if response.hasFailed():
        raise NetworkRequestError(response, f"Failed to create sample from \"{samplePath}\"")

    return sampleType.decode(response.getJson(dict))


def _encryptedSampleImport(sampleType: Type[SampleType], sampleName: str, samplePath: Path, datasetId: int, key: bytes) -> SampleType:
    with ExitStack() as stack:
        if isArchive(samplePath):
            archivePath = samplePath
        else:
            archivePath = stack.enter_context(folder_manager.tempFile())
            archive(samplePath, archivePath)

        encryptedPath = stack.enter_context(folder_manager.tempFile())
        aes.encryptFile(key, archivePath, encryptedPath)

        return _chunkSampleImport(sampleType, sampleName, encryptedPath, datasetId)

    # ExitStack is marked as something that can "swallow" exceptions raised
    # in "with" block. In this case ExitStack does not enter any context
    # which can swallow the exceptions so this is not reachable, but mypy
    # is complaning about this and we silence it by raising an exception here
    raise RuntimeError("Unreachable statement was reached.")


class NetworkDataset(Generic[SampleType], Dataset[SampleType], NetworkObject, ABC):

    """
        Represents the base class for all Dataset classes which are
        comunicating with Coretex.ai

        Properties
        ----------
        createdOn : datetime
            creation date of dataset
        createdById : str
            id of created dataset id
        isLocked : bool
            availabilty of dataset for modifications
    """

    projectId: int
    createdOn: datetime
    createdById: str
    isLocked: bool
    isEncrypted: bool
    meta: Optional[Dict[str, Any]]

    def __init__(self, sampleType: Type[SampleType]) -> None:
        self._sampleType = sampleType

    @property
    def path(self) -> Path:
        """
            Retrieves path of dataset

            Returns
            -------
            Path -> path of dataset
        """

        return folder_manager.datasetsFolder / str(self.id)

    # Codable overrides

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()

        descriptors["projectId"] = KeyDescriptor("project_id")
        descriptors["samples"] = KeyDescriptor("sessions", NetworkSample, list)

        return descriptors

    # NetworkObject overrides

    @classmethod
    def _endpoint(cls) -> str:
        return "dataset"

    @classmethod
    def fetchById(cls, objectId: int, **kwargs: Any) -> Self:
        if "include_sessions" not in kwargs:
            kwargs["include_sessions"] = 1

        return super().fetchById(objectId, **kwargs)

    @classmethod
    def fetchAll(cls, **kwargs: Any) -> List[Self]:
        if "include_sessions" not in kwargs:
            kwargs["include_sessions"] = 1

        return super().fetchAll(**kwargs)

    @classmethod
    def fetchCachedDataset(cls, dependencies: List[str]) -> Self:
        """
            Fetches cached dataset if it exists

            Parameters
            ----------
            dependencies : List[str]
                Parameters on which the cached dataset depends

            Returns
            -------
            Self -> Fetched dataset object

            Raises
            ------
            ValueError -> If dataset doesn't exist
        """

        return super().fetchOne(
            name = _hashDependencies(dependencies),
            include_sessions = 1
        )

    # Dataset methods

    @classmethod
    def createDataset(
        cls,
        name: str,
        projectId: int,
        meta: Optional[Dict[str, Any]] = None
    ) -> Self:

        """
            Creates a new dataset with the provided name and type

            Parameters
            ----------
            name : str
                dataset name
            projectId : int
                project for which the dataset will be created

            Returns
            -------
            The created dataset object or None if creation failed

            Raises
            ------
            NetworkRequestError -> If dataset creation failed

            Example
            -------
            >>> from coretex import NetworkDataset
            \b
            >>> dummyDataset = NetworkDataset.createDataset("dummyDataset", 123)
        """

        return cls.create(
            name = name,
            project_id = projectId,
            meta = meta
        )

    @classmethod
    def generateCacheName(cls, prefix: str, dependencies: List[str]) -> str:
        """
            Generated dataset name based on the dependencies

            Parameters
            ----------
            prefix : str
                prefix to which the dependency hash will be appended
            dependencies : List[str]
                parameters which affect the contents of the cache

            Returns
            -------
            str -> prefix with hash generated based on dependencies appended
        """

        if MAX_DATASET_NAME_LENGTH - len(prefix) < 8:
            raise ValueError(f"Dataset prefix \"{prefix}\" is too long. Max allowed size is \"{MAX_DATASET_NAME_LENGTH - 8}\".")

        suffix = _hashDependencies(dependencies)
        name = f"{prefix} - {suffix}"

        if len(name) > MAX_DATASET_NAME_LENGTH:
            name = name[:MAX_DATASET_NAME_LENGTH]

        return name

    @classmethod
    def createCacheDataset(cls, prefix: str, dependencies: List[str], projectId: int) -> Self:
        """
            Creates a dataset used for caching results of tasks
            Used to avoid repeating expensive and long calculations

            Parameters
            ----------
            prefix : str
                prefix of the cache dataset
            dependencies : List[str]
                parameters which affect the contents of the cache
            projectId : int
                project for which the dataset will be created
        """

        dataset = cls.createDataset(cls.generateCacheName(prefix, dependencies), projectId)
        if dataset is None:
            raise ValueError(f"Failed to create cache dataset with prefix \"{prefix}\"")

        return dataset

    def finalize(self) -> bool:
        """
            Finalizes state of Coretex dataset

            Example
            -------
            >>> from coretex import CustomDataset
            \b
            >>> dummyDataset = CustomDataset.createDataset("dummyDataset", 123)
            >>> dummyDataset.finalize()
        """

        return self.update(name = self.name, state = DatasetState.final)

    def _linkSamplePath(self, samplePath: Path) -> None:
        linkPath = self.path / samplePath.name
        if linkPath.exists():
            linkPath.unlink()

        samplePath.link_to(linkPath)

    def download(self, decrypt: bool = True, ignoreCache: bool = False) -> None:
        """
            Downloads dataset from Coretex

            Parameters
            ----------
            ignoreCache : bool
                if dataset is already downloaded and ignoreCache
                is True it will be downloaded again (not required)

            Example
            -------
            >>> from coretex import NetworkDataset
            \b
            >>> dummyDataset = NetworkDataset.fetchById(1023)
            >>> dummyDataset.download()
        """

        self.path.mkdir(exist_ok = True)

        def sampleDownloader(sample: SampleType) -> None:
            sample.download(decrypt, ignoreCache)

            if sample.downloadPath.exists():
                self._linkSamplePath(sample.downloadPath)

            if sample.zipPath.exists():
                self._linkSamplePath(sample.zipPath)

            logging.getLogger("coretexpylib").info(f"\tDownloaded \"{sample.name}\"")

        processor = MultithreadedDataProcessor(
            self.samples,
            sampleDownloader,
            message = f"Downloading dataset \"{self.name}\"..."
        )

        processor.process()

    def rename(self, name: str) -> bool:
        success = self.update(name = name)

        if success:
            return super().rename(name)

        return success

    @abstractmethod
    def _uploadSample(self, samplePath: Path, sampleName: str, **metadata: Any) -> SampleType:
        # Override in data specific classes (ImageDataset, SequenceDataset, etc...)
        # to implement a specific way of uploading samples
        pass

    def add(self, samplePath: Union[Path, str], sampleName: Optional[str] = None, **metadata: Any) -> SampleType:
        """
            Uploads the provided archive (.zip, .tar.gz) as Sample to
            Coretex.ai as a part of this Dataset.

            Parametrs
            ---------
            path : Union[Path, str]
                path to data which will be uploaded

            Returns
            -------
            SampleType -> created Sample
        """

        if isinstance(samplePath, str):
            samplePath = Path(samplePath)

        if sampleName is None:
            sampleName = samplePath.stem

        if self.isEncrypted:
            sample = _encryptedSampleImport(self._sampleType, sampleName, samplePath, self.id, getProjectKey(self.projectId))
        else:
            sample = self._uploadSample(samplePath, sampleName, **metadata)

        # Append the newly created sample to the list of samples
        self.samples.append(sample)

        return sample



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Union
from pathlib import Path

import os

from .utils import compressGzip, createSample, getDemuxSamples, getDenoisedSamples, \
    getFastqDPSamples, getFastqMPSamples, getImportedSamples, getMetadata, getPhylogeneticTreeSamples, \
    isDemultiplexedSample, isDenoisedSample, isFastqDPSample, isFastqMPSample, \
    isImportedSample, isPhylogeneticTreeSample, sampleNumber, isPairedEnd

from ...utils import command


def toolsImport(
    sequenceType: str,
    inputPath: str,
    outputPath: str,
    inputFormat: Optional[str] = None
) -> None:

    args = [
        "qiime", "tools", "import",
        "--type", sequenceType,
        "--input-path", str(inputPath),
        "--output-path", str(outputPath)
    ]

    if inputFormat is not None:
        args.extend([
            "--input-format" , inputFormat
        ])

    command(args)


def demuxEmpSingle(
    sequencesPath: str,
    barcodesPath: str,
    barcodesColumn: str,
    perSampleSequences: str,
    errorCorretctionDetailsPath: str
) -> None:

    command([
        "qiime", "demux", "emp-single",
        "--i-seqs", sequencesPath,
        "--m-barcodes-file", barcodesPath,
        "--m-barcodes-column", barcodesColumn,
        "--o-per-sample-sequences", perSampleSequences,
        "--o-error-correction-details", errorCorretctionDetailsPath,
        "--verbose"
    ])


def demuxEmpPaired(
    sequencesPath: Union[str, Path],
    barcodesPath: Union[str, Path],
    barcodesColumn: str,
    perSampleSequences: Union[str, Path],
    errorCorretctionDetailsPath: Union[str, Path]
) -> None:
    """
        Wrapper for QIIME's demux emp-paired command, which performs demultiplexing of multiplexed
        sequences in the Earth Microbiome Project amplicon sequencing standard format
        http://www.earthmicrobiome.org/protocols-and-standards/.

        Parameters
        ----------
        sequencesPath : Union[str, Path]
            Path to the paired-end demultiplexed sequences to be denoised
        barcodesPath : Union[str, Path]
            Output path to the resulting feature sequences
        perSampleSequences : Union[str, Path]
            Output path to the resulting feature table
        errorCorretctionDetailsPath : Union[str, Path]
            Output path to the statistics of the denoising
    """

    if isinstance(sequencesPath, Path):
        sequencesPath = str(sequencesPath)

    if isinstance(barcodesPath, Path):
        barcodesPath = str(barcodesPath)

    if isinstance(perSampleSequences, Path):
        perSampleSequences = str(perSampleSequences)

    if isinstance(errorCorretctionDetailsPath, Path):
        errorCorretctionDetailsPath = str(errorCorretctionDetailsPath)

    command([
        "qiime", "demux", "emp-paired",
        "--i-seqs", sequencesPath,
        "--m-barcodes-file", barcodesPath,
        "--m-barcodes-column", barcodesColumn,
        "--o-per-sample-sequences", perSampleSequences,
        "--o-error-correction-details", errorCorretctionDetailsPath,
        "--verbose"
    ])


def demuxSummarize(dataPath: str, visualizationPath: str) -> None:
    command([
        "qiime", "demux", "summarize",
        "--i-data", dataPath,
        "--o-visualization", visualizationPath
    ])


def dada2DenoiseSingle(
    inputPath: Union[str, Path],
    trimLeft: int,
    truncLen: int,
    representativeSequencesPath: Union[str, Path],
    tablePath: Union[str, Path],
    denoisingStatsPath: Union[str, Path],
    threads: Optional[int] = -1
) -> None:
    """
        Wrapper for QIIME's DADA2 denoise-single, which performs denoising on the input paired-end reads.

        Parameters
        ----------
        inputPath : Union[str, Path]
            Path to the paired-end demultiplexed sequences to be denoised
        trimLeft : int
            Position at which read sequences should be trimmed due to low quality.
            This trims the 5' end of the input sequences, which will be the bases that were
            sequenced in the first cycles
        truncLen : int
            Position at which read sequences should be truncated due to decrease in quality.
            This truncates the 3' end of the input sequences, which will be the bases that
            were sequenced in the last cycles. Reads that are shorter than this value will be discarded.
            After this parameter is applied there must still be at least a 20 nucleotide overlap
            between the forward and reverse reads. If 0 is provided, no truncation or length
            filtering will be performed
        representativeSequencesPath : Union[str, Path]
            Output path to the resulting feature sequences. Each feature in the feature table will
            be represented by exactly one sequence, and these sequences will be the
            joined paired-end sequences
        tablePath : Union[str, Path]
            Output path to the resulting feature table
        denoisingStatsPath : Union[str, Path]
            Output path to the statistics of the denoising
        threads : Optional[int]
            Number of threads to use for the process. If None, multithreading will not be used.
            Set to -1 (defaul) to use a thread for each CPU core
    """

    if isinstance(inputPath, Path):
        inputPath = str(inputPath)

    if isinstance(representativeSequencesPath, Path):
        representativeSequencesPath = str(representativeSequencesPath)

    if isinstance(tablePath, Path):
        tablePath = str(tablePath)

    if isinstance(denoisingStatsPath, Path):
        denoisingStatsPath = str(denoisingStatsPath)

    args = [
        "qiime", "dada2", "denoise-single",
        "--i-demultiplexed-seqs", inputPath,
        "--p-trim-left", str(trimLeft),
        "--p-trunc-len", str(truncLen),
        "--o-representative-sequences", representativeSequencesPath,
        "--o-table", tablePath,
        "--o-denoising-stats", denoisingStatsPath
    ]

    if threads == -1:
        args.extend(["--p-n-threads", "0"])
    elif threads is not None:
        args.extend(["--p-n-threads", str()])

    command(args)


def dada2DenoisePaired(
    inputPath: Union[str, Path],
    trimLeftF: int,
    trimLeftR: int,
    truncLenF: int,
    truncLenR: int,
    representativeSequencesPath: Union[str, Path],
    tablePath: Union[str, Path],
    denoisingStatsPath: Union[str, Path],
    threads: Optional[int] = -1
) -> None:
    """
        Wrapper for QIIME's DADA2 denoise-paired, which performs denoising on the input paired-end reads.

        Parameters
        ----------
        inputPath : Union[str, Path]
            Path to the paired-end demultiplexed sequences to be denoised
        trimLeftF : int
            Position at which forward read sequences should be trimmed due to low quality.
            This trims the 5' end of the input sequences, which will be the bases that were
            sequenced in the first cycles
        trimLeftR : int
            Position at which reverse read sequences should be trimmed due to low quality.
            This trims the 5' end of the input sequences, which will be the bases that were
            sequenced in the first cycles
        truncLenF : int
            Position at which forward read sequences should be truncated due to decrease in quality.
            This truncates the 3' end of the input sequences, which will be the bases that
            were sequenced in the last cycles. Reads that are shorter than this value will be discarded.
            After this parameter is applied there must still be at least a 20 nucleotide overlap
            between the forward and reverse reads. If 0 is provided, no truncation or length
            filtering will be performed
        truncLenR : int
            Position at which reverse read sequences should be truncated due to decrease in quality.
            This truncates the 3' end of the input sequences, which will be the bases that
            were sequenced in the last cycles. Reads that are shorter than this value will be discarded.
            After this parameter is applied there must still be at least a 20 nucleotide overlap
            between the forward and reverse reads. If 0 is provided, no truncation or length
            filtering will be performed
        representativeSequencesPath : Union[str, Path]
            Output path to the resulting feature sequences. Each feature in the feature table will
            be represented by exactly one sequence, and these sequences will be the
            joined paired-end sequences
        tablePath : Union[str, Path]
            Output path to the resulting feature table
        denoisingStatsPath : Union[str, Path]
            Output path to the statistics of the denoising
        threads : Optional[int]
            Number of threads to use for the process. If None, multithreading will not be used.
            Set to -1 (defaul) to use a thread for each CPU core
    """

    if isinstance(inputPath, Path):
        inputPath = str(inputPath)

    if isinstance(representativeSequencesPath, Path):
        representativeSequencesPath = str(representativeSequencesPath)

    if isinstance(tablePath, Path):
        tablePath = str(tablePath)

    if isinstance(denoisingStatsPath, Path):
        denoisingStatsPath = str(denoisingStatsPath)

    args = [
        "qiime", "dada2", "denoise-paired",
        "--i-demultiplexed-seqs", inputPath,
        "--p-trim-left-f", str(trimLeftF),
        "--p-trim-left-r", str(trimLeftR),
        "--p-trunc-len-f", str(truncLenF),
        "--p-trunc-len-r", str(truncLenR),
        "--o-representative-sequences", representativeSequencesPath,
        "--o-table", tablePath,
        "--o-denoising-stats", denoisingStatsPath
    ]

    if threads == -1:
        args.extend(["--p-n-threads", "0"])
    elif threads is not None:
        args.extend(["--p-n-threads", str(threads)])

    command(args)


def metadataTabulate(inputFile: str, visualizationPath: str) -> None:
    command([
        "qiime", "metadata", "tabulate",
        "--m-input-file", inputFile,
        "--o-visualization", visualizationPath
    ])


def featureTableSummarize(inputPath: str, visualizationPath: str, metadataPath: str) -> None:
    command([
        "qiime", "feature-table", "summarize",
        "--i-table", inputPath,
        "--o-visualization", visualizationPath,
        "--m-sample-metadata-file", metadataPath
    ])


def featureTableTabulateSeqs(inputPath: str, visualizationPath: str) -> None:
    command([
        "qiime", "feature-table", "tabulate-seqs",
        "--i-data", inputPath,
        "--o-visualization", visualizationPath
    ])


def phylogenyAlignToTreeMafftFasttree(
    sequencesPath: str,
    aligmentPath: str,
    maskedAligmentPath: str,
    unrootedTreePath: str,
    rootedTreePath: str,
    threads: Optional[int] = -1
) -> None:

    args = [
        "qiime", "phylogeny", "align-to-tree-mafft-fasttree",
        "--i-sequences", sequencesPath,
        "--o-alignment", aligmentPath,
        "--o-masked-alignment", maskedAligmentPath,
        "--o-tree", unrootedTreePath,
        "--o-rooted-tree", rootedTreePath
    ]

    if threads == -1:
        args.extend(["--p-n-threads", "auto"])
    elif threads is not None:
        args.extend(["--p-n-threads", str(threads)])

    command(args)


def diversityCoreMetricsPhylogenetic(
    phlogenyPath: str,
    tablePath: str,
    samplingDepth: int,
    metadataPath: str,
    outputDir: str,
    threads: Optional[int] = -1
) -> None:

    args = [
        "qiime", "diversity", "core-metrics-phylogenetic",
        "--i-phylogeny", phlogenyPath,
        "--i-table", tablePath,
        "--p-sampling-depth", str(samplingDepth),
        "--m-metadata-file", metadataPath,
        "--output-dir", outputDir
    ]

    if threads == -1:
        args.extend(["--p-n-jobs-or-threads", "auto"])
    elif threads is not None:
        args.extend(["--p-n-jobs-or-threads", str(threads)])

    command(args)


def diversityAlphaGroupSignificance(
    alphaDiversityPath: str,
    metadataPath: str,
    visualizationPath: str
) -> None:

    command([
        "qiime", "diversity", "alpha-group-significance",
        "--i-alpha-diversity", alphaDiversityPath,
        "--m-metadata-file", metadataPath,
        "--o-visualization", visualizationPath
    ])


def diversityBetaGroupSignificance(
    distanceMatrixPath: str,
    metadataPath: str,
    metadataColumn: str,
    visualizationPath: str,
    pairwise: bool
) -> None:

    command([
        "qiime", "diversity", "beta-group-significance",
        "--i-distance-matrix", distanceMatrixPath,
        "--m-metadata-file", metadataPath,
        "--m-metadata-column", metadataColumn,
        "--o-visualization", visualizationPath,
        "--p-pairwise" if pairwise else ""
    ])


def emperorPlot(
    pcoaPath: str,
    metadataPath: str,
    customAxes: str,
    visualizationPath: str
) -> None:

    command([
        "qiime", "emperor", "plot",
        "--i-pcoa", pcoaPath,
        "--m-metadata-file", metadataPath,
        "--p-custom-axes", customAxes,
        "--o-visualization", visualizationPath
    ])


def diversityAlphaRarefaction(
    tablePath: str,
    phylogenyPath: str,
    maxDepth: int,
    metadataPath: str,
    visualizationPath: str
) -> None:

    command([
        "qiime", "diversity", "alpha-rarefaction",
        "--i-table", tablePath,
        "--i-phylogeny", phylogenyPath,
        "--p-max-depth", str(maxDepth),
        "--m-metadata-file", metadataPath,
        "--o-visualization", visualizationPath
    ])


def featureClassifierClassifySklearn(
    classifierPath: str,
    readsPath: str,
    classificationPath: str,
    threads: Optional[int] = -1
) -> None:

    args = [
        "qiime", "feature-classifier", "classify-sklearn",
        "--i-classifier", classifierPath,
        "--i-reads", readsPath,
        "--o-classification", classificationPath
    ]

    if threads == -1:
        args.extend(["--p-n-jobs", "-1"])
    elif threads is not None:
        args.extend(["--p-n-jobs", str(threads)])

    command(args)


def taxaBarplot(
    tablePath: str,
    taxonomyPath: str,
    metadataPath: str,
    visualizationPath: str
) -> None:

    command([
        "qiime", "taxa", "barplot",
        "--i-table", tablePath,
        "--i-taxonomy", taxonomyPath,
        "--m-metadata-file", metadataPath,
        "--o-visualization", visualizationPath
    ])


def featureTableFilterSamples(
    tablePath: str,
    metadataPath: str,
    where: str,
    filteredTablePath: str
) -> None:

    command([
        "qiime", "feature-table", "filter-samples",
        "--i-table", tablePath,
        "--m-metadata-file", metadataPath,
        "--p-where", where,
        "--o-filtered-table", filteredTablePath
    ])


def compositionAddPseudocount(tablePath: str, compositionTablePath: str) -> None:
    command([
        "qiime", "composition", "add-pseudocount",
        "--i-table", tablePath,
        "--o-composition-table", compositionTablePath
    ])


def compositionAncom(
    tablePath: str,
    metadataPath: str,
    metadataColumn: str,
    visualizationPath: str
) -> None:

    command([
        "qiime", "composition", "ancom",
        "--i-table", tablePath,
        "--m-metadata-file", metadataPath,
        "--m-metadata-column", metadataColumn,
        "--o-visualization", visualizationPath
    ])


def taxaCollapse(
    tablePath: str,
    taxonomyPath: str,
    level: int,
    collapsedTablePath: str
) -> None:

    command([
        "qiime", "taxa", "collapse",
        "--i-table", tablePath,
        "--i-taxonomy", taxonomyPath,
        "--p-level", str(level),
        "--o-collapsed-table", collapsedTablePath
    ])

def vsearchClusterDeNovo(
    tablePath: Union[str, Path],
    representativeSequencesPath: Union[str, Path],
    percIdentity: float,
    clusteredTablePath: Union[str, Path],
    clusteredSequencesPath: Union[str, Path],
    threads: Optional[int] = -1
) -> None:
    """
        Wrapper for QIIME2's vsearch de novo clusteing command.

        Parameters
        ----------
        tablePath : Union[str, Path]
            Path to the feature table generated by DADA2 denoise
        representativeSequencesPath : Union[str, Path]
            Path to the represenative sequences generated by DADA2 denoise
        percIdentity : float
            Percent identiy threshold for the OTU clustering
        clusteredTablePath : Union[str, Path]
            Path to the output clustered feature table
        clusteredSequencesPath : Union[str, Path]
            Path to the output clustered sequences
        multithreading : bool
            Number of threads to use for the process. If None, multithreading will not be used.
            Set to -1 (defaul) to use a thread for each CPU core
    """

    if isinstance(tablePath, Path):
        tablePath = str(tablePath)

    if isinstance(representativeSequencesPath, Path):
        representativeSequencesPath = str(representativeSequencesPath)

    if isinstance(clusteredTablePath, Path):
        clusteredTablePath = str(clusteredTablePath)

    if isinstance(clusteredSequencesPath, Path):
        clusteredSequencesPath = str(clusteredSequencesPath)

    args = [
        "qiime", "vsearch", "cluster-features-de-novo",
        "--i-table", tablePath,
        "--i-sequences", representativeSequencesPath,
        "--p-perc-identity", str(percIdentity),
        "--o-clustered-table", clusteredTablePath,
        "--o-clustered-sequences", clusteredSequencesPath
    ]

    if threads == -1:
        args.extend(["--p-threads", "0"])
    elif threads is not None:
        args.extend(["--p-threads", str(threads)])

    command(args)


def vsearchClusterClosedReference(
    tablePath: Union[str, Path],
    representativeSequencesPath: Union[str, Path],
    referenceSequencesPath: Union[str, Path],
    percIdentity: float,
    clusteredTablePath: Union[str, Path],
    clusteredSequencesPath: Union[str, Path],
    unmatchedSequencesPath: Union[str, Path],
    threads: Optional[int] = -1
) -> None:
    """
        Wrapper for QIIME2's vsearch closed reference clusteing command.

        Parameters
        ----------
        tablePath : Union[str, Path]
            Path to the feature table generated by DADA2 denoise
        representativeSequencesPath : Union[str, Path]
            Path to the represenative sequences generated by DADA2 denoise
        referenceSequencesPath : Union[str, Path]
            Path to reference OTU sequences
        percIdentity : float
            Percent identiy threshold for the OTU clustering
        clusteredTablePath : Union[str, Path]
            Path to the output clustered feature table
        clusteredSequencesPath : Union[str, Path]
            Path to the output clustered sequences
        unmatchedSequencesPath : Union[str, Path]
            Path to the output unmatched sequences
        multithreading : bool
            Number of threads to use for the process. If None, multithreading will not be used.
            Set to -1 (defaul) to use a thread for each CPU core
    """

    if isinstance(tablePath, Path):
        tablePath = str(tablePath)

    if isinstance(representativeSequencesPath, Path):
        representativeSequencesPath = str(representativeSequencesPath)

    if isinstance(referenceSequencesPath, Path):
        referenceSequencesPath = str(referenceSequencesPath)

    if isinstance(clusteredTablePath, Path):
        clusteredTablePath = str(clusteredTablePath)

    if isinstance(clusteredSequencesPath, Path):
        clusteredSequencesPath = str(clusteredSequencesPath)

    if isinstance(unmatchedSequencesPath, Path):
        unmatchedSequencesPath = str(unmatchedSequencesPath)

    args = [
        "qiime", "vsearch", "cluster-features-closed-reference",
        "--i-table", tablePath,
        "--i-sequences", representativeSequencesPath,
        "--i-reference-sequences", referenceSequencesPath,
        "--p-perc-identity", str(percIdentity),
        "--o-clustered-table", clusteredTablePath,
        "--o-clustered-sequences", clusteredSequencesPath,
        "--o-unmatched-sequences", unmatchedSequencesPath
    ]

    if threads == -1:
        args.extend(["--p-threads", "0"])
    elif threads is not None:
        args.extend(["--p-threads", str(threads)])

    command(args)


def vsearchClusterOpenReference(
    tablePath: Union[str, Path],
    representativeSequencesPath: Union[str, Path],
    referenceSequencesPath: Union[str, Path],
    percIdentity: float,
    clusteredTablePath: Union[str, Path],
    clusteredSequencesPath: Union[str, Path],
    newReferenceSequencesPath: Union[str, Path],
    threads: Optional[int] = -1
) -> None:
    """
        Wrapper for QIIME2's vsearch open reference clusteing command.

        Parameters
        ----------
        tablePath : Union[str, Path]
            Path to the feature table generated by DADA2 denoise
        representativeSequencesPath : Union[str, Path]
            Path to the represenative sequences generated by DADA2 denoise
        referenceSequencesPath : Union[str, Path]
            Path to reference OTU sequences
        percIdentity : float
            Percent identiy threshold for the OTU clustering
        clusteredTablePath : Union[str, Path]
            Path to the output clustered feature table
        clusteredSequencesPath : Union[str, Path]
            Path to the output clustered sequences
        newReferenceSequencesPath : Union[str, Path]
            Path to the output new reference sequences
        multithreading : bool
            Number of threads to use for the process. If None, multithreading will not be used.
            Set to -1 (defaul) to use a thread for each CPU core
    """

    if isinstance(tablePath, Path):
        tablePath = str(tablePath)

    if isinstance(representativeSequencesPath, Path):
        representativeSequencesPath = str(representativeSequencesPath)

    if isinstance(referenceSequencesPath, Path):
        referenceSequencesPath = str(referenceSequencesPath)

    if isinstance(clusteredTablePath, Path):
        clusteredTablePath = str(clusteredTablePath)

    if isinstance(clusteredSequencesPath, Path):
        clusteredSequencesPath = str(clusteredSequencesPath)

    if isinstance(newReferenceSequencesPath, Path):
        newReferenceSequencesPath = str(newReferenceSequencesPath)

    args = [
        "qiime", "vsearch", "cluster-features-open-reference",
        "--i-table", tablePath,
        "--i-sequences", representativeSequencesPath,
        "--i-reference-sequences", referenceSequencesPath,
        "--p-perc-identity", str(percIdentity),
        "--o-clustered-table", clusteredTablePath,
        "--o-clustered-sequences", clusteredSequencesPath,
        "--o-new-reference-sequences", newReferenceSequencesPath
    ]

    if threads == -1:
        args.extend(["--p-threads", "0"])
    elif threads is not None:
        args.extend(["--p-threads", str(threads)])

    command(args)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List
from pathlib import Path

from ...utils import command


def getExisting() -> List[str]:
    _, output, error = command(["crontab", "-l"], ignoreStdout = True, ignoreStderr = True, check = False)
    if error is not None and "no crontab for" in error:
        return []
    if output is not None:
        return [line.strip() for line in output.split("\n") if line.strip()]

    raise ValueError("\"crontab\" is not installed. To enable automatic updates please install \"crontab\"")


def jobExists(script: str) -> bool:
    existingLines = getExisting()
    return any(line.endswith(script) for line in existingLines)


def scheduleJob(configDir: Path, script: str) -> None:
    if jobExists(script):
        return

    existingLines = getExisting()
    cronEntry = f"*/30 * * * * {configDir / script}\n"
    existingLines.append(cronEntry)

    tempCronFilePath = configDir / "temp.cron"
    with tempCronFilePath.open("w") as tempCronFile:
        tempCronFile.write("\n".join(existingLines))

    command(["crontab", str(tempCronFilePath)])
    tempCronFilePath.unlink()



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from multiprocessing.connection import Connection

import logging

from ... import folder_manager
from ...utils import createFileHandler


def sendSuccess(conn: Connection, message: str) -> None:
    conn.send({
        "code": 0,
        "message": message
    })


def sendFailure(conn: Connection, message: str) -> None:
    conn.send({
        "code": 1,
        "message": message
    })


def initializeLogger(taskRunId: int) -> None:
    formatter = logging.Formatter(
        fmt = "%(asctime)s %(levelname)s: %(message)s",
        datefmt = "%Y-%m-%d %H:%M:%S",
        style = "%",
    )

    workerLogPath = folder_manager.getRunLogsDir(taskRunId) / "worker.log"
    fileHandler = createFileHandler(workerLogPath)

    fileHandler.setLevel(logging.DEBUG)
    fileHandler.setFormatter(formatter)

    logging.basicConfig(
        level = logging.NOTSET,
        force = True,
        handlers = [fileHandler]
    )



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Tuple

from .type import SecretType
from .secret import Secret
from ...codable import KeyDescriptor


class AWSSecret(Secret):

    """
        Represents AWS Secret entity from Coretex.ai

        Properties
        ----------
        key : str
            AWS Access Key ID
        value : str
            AWS Secret Access Key
    """

    key: str
    value: str

    def __init__(self) -> None:
        super().__init__(SecretType.aws)

    @classmethod
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()
        descriptors["value"] = KeyDescriptor("secret")

        return descriptors

    def _encryptedFields(self) -> Tuple[str, ...]:
        return ("key", "value")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Any
from typing_extensions import override
from pathlib import Path

from .base import BaseCustomDataset
from ..network_dataset import NetworkDataset, _chunkSampleImport
from ...sample import CustomSample
from ....codable import KeyDescriptor


class CustomDataset(BaseCustomDataset, NetworkDataset[CustomSample]):

    """
        Custom Dataset class which is used for Other Task
        Represents the collection of archived samples
    """

    @override
    def __init__(self) -> None:
        super().__init__(CustomSample)

    @classmethod
    @override
    def _keyDescriptors(cls) -> Dict[str, KeyDescriptor]:
        descriptors = super()._keyDescriptors()
        descriptors["samples"] = KeyDescriptor("sessions", CustomSample, list)

        return descriptors

    @override
    def _uploadSample(self, samplePath: Path, sampleName: str, **metadata: Any) -> CustomSample:
        return _chunkSampleImport(self._sampleType, sampleName, samplePath, self.id)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Any, Dict, List, Union, Tuple, BinaryIO

import io
import logging

from .network_response import NetworkResponse


RequestBodyType = Dict[str, Any]
RequestFormType = List[Tuple[str, Tuple[str, Union[bytes, BinaryIO], str]]]


def logFilesData(files: Optional[RequestFormType]) -> List[Dict[str, Any]]:
    if files is None:
        return []

    debugFilesData: List[Dict[str, Any]] = []

    for paramName, (fileName, fileData, mimeType) in files:
        if isinstance(fileData, bytes):
            fileSize = len(fileData)
        else:
            fileSize = fileData.seek(0, io.SEEK_END)
            fileData.seek(0)

        debugFilesData.append({
            "paramName": paramName,
            "fileName": fileName,
            "fileSize": fileSize,  # Dump file size instead of file data because file data can be in GBs
            "mimeType": mimeType
        })

    return debugFilesData


def logRequestFailure(endpoint: str, response: NetworkResponse) -> None:
    if not response.hasFailed():
        raise ValueError(f">> [Coretex] Invalid response status code: \"{response.statusCode}\"")

    logging.getLogger("coretexpylib").debug(f">> [Coretex] Request to \"{endpoint}\" failed with status code: {response.statusCode}")

    try:
        responseJson = response.getJson(dict)

        message = responseJson.get("message")
        if message is not None:
            logging.getLogger("coretexpylib").debug(f"\tResponse: {message}")
        else:
            logging.getLogger("coretexpylib").debug(f"\tResponse: {responseJson}")
    except (ValueError, TypeError):
        logging.getLogger("coretexpylib").debug(f"\tResponse: {response.getContent()!r}")



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Any
from dataclasses import dataclass
from datetime import datetime, timezone

import logging

from .ui import clickPrompt, errorEcho, progressEcho
from ...utils import decodeDate
from ...networking import networkManager, NetworkResponse, NetworkRequestError
from ...configuration import loadConfig, saveConfig


@dataclass
class LoginInfo:

    username: str
    password: str
    token: str
    tokenExpirationDate: str
    refreshToken: str
    refreshTokenExpirationDate: str


def authenticateUser(username: str, password: str) -> NetworkResponse:
    response = networkManager.authenticate(username, password, False)

    if response.hasFailed():
        if response.statusCode >= 500:
            raise NetworkRequestError(response, "Something went wrong, please try again later.")

        if response.statusCode >= 400:
            raise NetworkRequestError(response, "User credentials invalid.")

    return response


def saveLoginData(loginInfo: LoginInfo, config: Dict[str, Any]) -> Dict[str, Any]:
    config["username"] = loginInfo.username
    config["password"] = loginInfo.password
    config["token"] = loginInfo.token
    config["tokenExpirationDate"] = loginInfo.tokenExpirationDate
    config["refreshToken"] = loginInfo.refreshToken
    config["refreshTokenExpirationDate"] = loginInfo.refreshTokenExpirationDate

    return config


def authenticate(retryCount: int = 0) -> LoginInfo:
    if retryCount >= 3:
        raise RuntimeError("Failed to authenticate. Terminating...")

    username = clickPrompt("Email", type = str)
    password = clickPrompt("Password", type = str, hide_input = True)

    progressEcho("Authenticating...")
    response = networkManager.authenticate(username, password, False)

    if response.hasFailed():
        errorEcho("Failed to authenticate. Please try again...")
        return authenticate(retryCount + 1)

    jsonResponse = response.getJson(dict)

    return LoginInfo(
        username,
        password,
        jsonResponse["token"],
        jsonResponse["expires_on"],
        jsonResponse["refresh_token"],
        jsonResponse["refresh_expires_on"]
    )


def initializeUserSession() -> None:
    config = loadConfig()

    if config.get("username") is None or config.get("password") is None:
        errorEcho("User configuration not found. Please authenticate with your credentials.")
        loginInfo = authenticate()
        config = saveLoginData(loginInfo, config)
    else:
        tokenExpirationDate = config.get("tokenExpirationDate")
        refreshTokenExpirationDate = config.get("refreshTokenExpirationDate")

        try:
            if tokenExpirationDate is not None and refreshTokenExpirationDate is not None:
                tokenExpirationDate = decodeDate(tokenExpirationDate)
                refreshTokenExpirationDate = decodeDate(refreshTokenExpirationDate)

                currentDate = datetime.utcnow().replace(tzinfo = timezone.utc)
                if currentDate < tokenExpirationDate:
                    return

                if currentDate < refreshTokenExpirationDate:
                    refreshToken = config["refreshToken"]
                    response = networkManager.authenticateWithRefreshToken(refreshToken)
                    if response.hasFailed():
                        if response.statusCode >= 500:
                            raise NetworkRequestError(response, "Something went wrong, please try again later.")

                        if response.statusCode >= 400:
                            response = authenticateUser(config["username"], config["password"])
                else:
                    response = authenticateUser(config["username"], config["password"])
            else:
                response = authenticateUser(config["username"], config["password"])
        except NetworkRequestError as ex:
            logging.getLogger("cli").debug(ex, exc_info = ex)

            if 400 <= ex.response.statusCode < 500:
                errorEcho("Authentification failed with configured credentials. Please try entering your credentials again.")
                authenticate()

        jsonResponse = response.getJson(dict)
        config["token"] = jsonResponse["token"]
        config["tokenExpirationDate"] = jsonResponse["expires_on"]
        config["refreshToken"] = jsonResponse.get("refresh_token")
        config["refreshTokenExpirationDate"] = jsonResponse.get("refresh_expires_on")

    saveConfig(config)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Type, Any, Optional, List, Generator
from pathlib import Path
from zipfile import ZipFile, ZIP_DEFLATED

import logging

import git

from ..dataset import Dataset, LocalDataset, NetworkDataset, LocalCustomDataset, \
    CustomDataset, LocalImageDataset, ImageDataset
from ..project import ProjectType
from ... import folder_manager


def getDatasetType(type_: ProjectType, isLocal: bool) -> Type[Dataset]:
    if type_ == ProjectType.other:
        if isLocal:
            return LocalCustomDataset

        return CustomDataset

    if type_ in [ProjectType.computerVision, ProjectType.imageSegmentation]:
        if isLocal:
            return LocalImageDataset

        return ImageDataset

    logging.getLogger("coretexpylib").debug(f">> [Coretex] ProjectType ({type_}) does not have a dataset type using CustomDataset")

    # Returning CustomDataset in case the type_ doesn't have it's dataset type
    if isLocal:
        return LocalCustomDataset

    return CustomDataset


def fetchDataset(datasetType: Type[Dataset], value: Any) -> Optional[Dataset]:
    if issubclass(datasetType, LocalDataset):
        return datasetType(value)  # type: ignore

    if issubclass(datasetType, NetworkDataset):
        return datasetType.fetchById(value)

    return None


def getSnapshotFiles(dirPath: Path, ignoredFiles: List[str]) -> List[Path]:
    snapshotFiles: List[Path] = []

    if dirPath.joinpath(".coretexignore").exists():
        return []

    for path in dirPath.iterdir():
        if path.is_dir():
            snapshotFiles.extend(getSnapshotFiles(path, ignoredFiles))
        elif str(path) not in ignoredFiles:
            snapshotFiles.append(path)

    return snapshotFiles


def getDefaultEntryPoint() -> Optional[str]:
    for defaultEntryPoint in [Path(".", "main.py"), Path(".", "main.r"), Path(".", "main.R")]:
        if defaultEntryPoint.exists():
            return defaultEntryPoint.name

    return None


def chunks(lst: List, n: int) -> Generator[List, None, None]:
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def createSnapshot() -> Path:
    entryPoint = getDefaultEntryPoint()
    if entryPoint is None or not Path(".", entryPoint).exists():
        raise FileNotFoundError(">> [Coretex] Entry point file not found")

    ignoredFiles: List[str] = []

    snapshotPath = folder_manager.temp / "snapshot.zip"
    with ZipFile(snapshotPath, "w", ZIP_DEFLATED) as snapshotArchive:
        repo = git.Repo(Path.cwd(), search_parent_directories = True)
        for paths in chunks(list(Path.cwd().rglob("*")), 256):
            ignoredFiles.extend(repo.ignored(*paths))

        if not Path(entryPoint).exists() or not Path("requirements.txt").exists():
            raise FileNotFoundError(f">> [Coretex] Required files \"{entryPoint}\" and \"requirements.txt\"")

        for path in getSnapshotFiles(Path.cwd(), ignoredFiles):
            snapshotArchive.write(path.relative_to(Path.cwd()))

    return snapshotPath



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Tuple, Union, BinaryIO
from typing_extensions import Self
from pathlib import Path
from contextlib import ExitStack

from ..utils import guessMimeType


class FileData:

    """
        Class which describes file which will be uploaded
        using NetworkManager.genericUpload function.
        To upload a file either its "filePath" or "fileBytes"
        must be set, otherwise it will raise an exception.
        "filePath" will upload the file from the specified path, while
        "fileBytes" will upload the file directly from the memory.
        If both parameters have value "fileBytes" will be used.

        Objects of this class should not be instantiated directly,
        use either "FileData.createFromPath" or "FileData.createFromBytes"
        to instantiate the object.

        Properties
        ----------
        parameterName : str
            Name of the form-data parameter
        fileName : str
            Name of the file which will be uploaded
        mimeType : str
            Mime type of the file which will be uploaded
        filePath : Optional[str]
            Path to the file which will be uploaded
        fileBytes : Optional[bytes]
            Bytes of the file which will be uploaded
    """

    def __init__(
        self,
        parameterName: str,
        fileName: str,
        mimeType: str,
        filePath: Optional[Path] = None,
        fileBytes: Optional[bytes] = None
    ) -> None:

        if filePath is None and fileBytes is None:
            raise ValueError(">> [Coretex] Either \"filePath\" or \"fileData\" have to provided for file upload. \"fileData\" will be used if both are provided")

        self.parameterName = parameterName
        self.fileName = fileName
        self.mimeType = mimeType
        self.filePath = filePath
        self.fileBytes = fileBytes

    @classmethod
    def createFromPath(
        cls,
        parameterName: str,
        filePath: Union[Path, str],
        fileName: Optional[str] = None,
        mimeType: Optional[str] = None
    ) -> Self:

        """
            Creates "FileData" object from the specified file path. Use
            this function if you want to upload a file directly from path.

            Parameters
            ----------
            parameterName : str
                Name of the form-data parameter
            filePath : Union[Path, str]
                Path to the file which will be uploaded
            fileName : Optional[str]
                Name of the file which will be uploaded, if None it will
                be extracted from the "filePath" parameter
            mimeType : Optional[str]
                Mime type of the file which will be uploaded, if None it will
                be guessed. If it is not possible to guess an exception will be
                raised. In that case provide the mime type manually.

            Returns
            -------
            Self -> on object which describes how a path should be uploaded from path

            Raises
            ------
            ValueError -> if "filePath" is not a valid file
        """

        if isinstance(filePath, str):
            filePath = Path(filePath)

        if not filePath.is_file():
            raise ValueError(">> [Coretex] \"filePath\" is not a valid file")

        if fileName is None:
            fileName = filePath.stem

        if mimeType is None:
            mimeType = guessMimeType(filePath)

        return cls(parameterName, fileName, mimeType, filePath = filePath)

    @classmethod
    def createFromBytes(
        cls,
        parameterName: str,
        fileBytes: bytes,
        fileName: str,
        mimeType: Optional[str] = None
    ) -> Self:

        """
            Creates "FileData" object from the provided bytes. Use this
            function if you want to upload a file directly from memory.

            Parameters
            ----------
            parameterName : str
                Name of the form-data parameter
            fileBytes : bytes
                Bytes of the file which will be uploaded
            fileName : str
                Name of the file which will be uploaded, if None it will
                be extracted from the "filePath" parameter
            mimeType : Optional[str]
                Mime type of the file which will be uploaded, if None it will
                be set to "application/octet-stream".

            Returns
            -------
            Self -> on object which describes how a path should be uploaded from memory
        """

        if mimeType is None:
            mimeType = "application/octet-stream"

        return cls(parameterName, fileName, mimeType, fileBytes = fileBytes)

    def __getFileData(self, exitStack: ExitStack) -> Union[bytes, BinaryIO]:
        if self.fileBytes is not None:
            return self.fileBytes

        if self.filePath is not None:
            return exitStack.enter_context(self.filePath.open("rb"))

        raise ValueError(">> [Coretex] Either \"filePath\" or \"fileData\" have to provided for file upload. \"fileData\" will be used if both are provided")

    def prepareForUpload(self, exitStack: ExitStack) -> Tuple[str, Tuple[str, Union[bytes, BinaryIO], str]]:
        """
            Converts the "FileData" object into a format which can be used
            by the requests library for uploading files.

            Parameters
            ----------
            exitStack : ExitStack
                Context stack which contains the context of files
                opened by the "FileData" object. Used to join multiple file
                contexts, so if one raises an exception all the files will
                properly get closed.

            Returns
            -------
            Tuple[str, Tuple[str, Any, str]] -> Format accepted by the requests
            library for uploading files.
        """

        return self.parameterName, (self.fileName, self.__getFileData(exitStack), self.mimeType)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Optional, Any, Tuple, List, TypeVar, Generic, Union
from abc import ABC, abstractmethod

import logging

from .parameter_type import ParameterType
from ...project import ProjectType


T = TypeVar("T")


class BaseParameter(ABC, Generic[T]):

    def __init__(
        self,
        name: str,
        description: str,
        value: Optional[T],
        dataType: Union[str, ParameterType],
        required: bool,
        type: int = 3
    ) -> None:

        if isinstance(dataType, str):
            dataType = ParameterType(dataType)

        self.name = name
        self.description = description
        self.value = value
        self.dataType = dataType
        self.required = required
        self.type_ = type

    @property
    @abstractmethod
    def types(self) -> List[type]:
        pass

    def makeExceptionMessage(self) -> str:
        expected = self.dataType.value
        received = self.generateTypeDescription()

        return f"Parameter \"{self.name}\" has invalid type. Expected \"{expected}\", got \"{received}\""

    def validate(self) -> Tuple[bool, Optional[str]]:
        if not self.required and self.value is None:
            return True, None

        # bool is a subclass of int, do not allow validation to pass if
        # we are looking for integer, but bool is received
        if isinstance(self.value, bool) and int in self.types and not bool in self.types:
            return False, None

        if not any(isinstance(self.value, dataType) for dataType in self.types):
            return False, None

        return True, None

    def generateTypeDescription(self) -> str:
        if not isinstance(self.value, list) or self.value is None:
            return type(self.value).__name__

        elementTypes = ", ".join({type(value).__name__ for value in self.value})
        return f"list[{elementTypes}]"

    def parseValue(self, type_: ProjectType) -> Optional[Any]:
        return self.value

    def overrideValue(self, value: Optional[Any]) -> Optional[Any]:
        return value

    def encode(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "description": self.description,
            "value": self.value,
            "data_type": self.dataType.value,
            "required": self.required
        }


def validateParameters(parameters: List[BaseParameter], verbose: bool = True) -> Dict[str, Any]:
    parameterValidationResults: Dict[str, bool] = {}

    for parameter in parameters:
        isValid, message = parameter.validate()
        if not isValid:
            if message is None:
                message = parameter.makeExceptionMessage()

            if verbose:
                logging.getLogger("coretexpylib").fatal(f">> [Coretex] {message}")

        parameterValidationResults[parameter.name] = isValid

    return parameterValidationResults



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

# from typing import Dict, Any, List, Tuple
# from pathlib import Path

# import wave
# import logging

# from deepspeech import Model

# import numpy as np

# from .token import Token
# from .utils import isTxtSample, getAudioFilePath
# from .transcription import Transcription
# from .text import loadTxtSample
# from ..coretex import CustomSample, CustomDataset
# from ..folder_management import FolderManager


# TranscriptionList = List[Tuple[CustomSample, Transcription]]


# def _instantiateModel(modelPath: Path, modelScorerPath: Path) -> Model:
#     model = Model(str(modelPath))
#     model.enableExternalScorer(str(modelScorerPath))

#     model.setScorerAlphaBeta(0.93, 1.18)  # 0.93 represents lm_alpha, and 1.18 represents lm_beta
#     model.setBeamWidth(100)  # 100 represents beam_width

#     return model


# def _readAudioFile(path: Path) -> bytes:
#     with wave.open(str(path), 'rb') as audioFile:
#         frames = audioFile.getnframes()
#         buffer = audioFile.readframes(frames)

#     return buffer


# def _hasCachedTranscription(dataset: CustomDataset, meta: Dict[str, Any]) -> bool:
#     fetchedDatasets = CustomDataset.fetchAll([
#         f"name={dataset.name}_cached",
#         "include_sessions=1"
#     ])

#     if fetchedDatasets != 1:
#         print("1", len(fetchedDatasets))
#         return False

#     cachedDataset = fetchedDatasets[0]
#     if cachedDataset.meta is None:
#         print("2", cachedDataset.meta)
#         return False

#     for key, value in meta.items():
#         if cachedDataset.meta.get(key) != value:
#             print("3", cachedDataset.meta.get(key), "!=", value)
#             return False

#     return True


# def _fetchCachedTranscription(dataset: CustomDataset, meta: Dict[str, Any]) -> TranscriptionList:
#     fetchedDatasets = CustomDataset.fetchAll([f"name={dataset.name}_cached"])
#     if fetchedDatasets != 1:
#         raise ValueError(">> [Coretex] Fetched more than one cached dataset")

#     cachedDataset = fetchedDatasets[0]
#     if cachedDataset.meta is None:
#         raise ValueError(">> [Coretex] Dataset.meta field is None")

#     for key, value in meta.items():
#         if cachedDataset.meta.get(key) != value:
#             raise ValueError(">> [Coretex] Dataset.meta field does not match parameters of TaskRun")

#     cachedDataset.download()

#     result: TranscriptionList = []

#     for sample in dataset.samples:
#         cachedSample = cachedDataset.getSample(f"{sample.name}_cached")
#         if cachedSample is None:
#             raise ValueError(f">> [Coretex] {sample.name} has invalid cache")

#         cachedSample.unzip()
#         folderContent = list(cachedSample.load().folderContent)

#         if len(folderContent) != 1:
#             raise ValueError(f">> [Coretex] {sample.name} has invalid cache")

#         result.append((sample, Transcription.load(folderContent[0])))

#     return result


# def _cacheTranscription(
#     dataset: CustomDataset,
#     transcriptions: TranscriptionList,
#     meta: Dict[str, Any]
# ) -> bool:

#     logging.info(">> [Coretex] Caching dataset transcription")

#     cachedDataset = CustomDataset.createDataset(f"{dataset.name}_cached", dataset.projectId, meta = meta)
#     if cachedDataset is None:
#         return False

#     try:
#         cachedSamplesPath = Path(FolderManager.instance().createTempFolder("cached_samples"))
#         for sample, transcription in transcriptions:
#             logging.info(f">> [Coretex] Caching sample: {sample.name}")

#             cachedSamplePath = transcription.save(cachedSamplesPath / f"{sample.name}.json")

#             cachedSample = CustomSample.createCustomSample(f"{sample.name}_cached", cachedDataset.id, str(cachedSamplePath))
#             if cachedSample is None:
#                 raise ValueError
#     except ValueError:
#         cachedDataset.delete()
#         return False

#     return True


# class AudioTranscriber:

#     """
#         Transcribes audio dataset into text, and tokens

#         Properties:
#         model: Model -> deepspeech model used for transcription
#         parameters: Dict[str, Any] -> parameters which affect model output,
#         these parameters are stored as metadata about cached dataset after transcription
#         has been performed
#     """

#     def __init__(self, modelPath: Path, modelScorerPath: Path, parameters: Dict[str, Any]):
#         self.model = _instantiateModel(modelPath, modelScorerPath)
#         self.parameters = parameters

#     def __transcribeSingle(self, sample: CustomSample, batchSize: int) -> Transcription:
#         logging.info(f">> [Coretex] Transcribing: {sample.name}")

#         sample.unzip()

#         if isTxtSample(sample):
#             return loadTxtSample(sample)

#         audioFilePath = getAudioFilePath(sample)
#         if audioFilePath is None:
#             raise ValueError(f">> [Coretex] {sample.name} does not contain a valid audio file")

#         stream = self.model.createStream()

#         buffer = _readAudioFile(audioFilePath)
#         offset = 0

#         while offset < len(buffer):
#             endOffset = offset + batchSize
#             chunk = buffer[offset:endOffset]

#             data16 = np.frombuffer(chunk, dtype = np.int16)
#             stream.feedAudioContent(data16)

#             text = stream.intermediateDecodeWithMetadata()
#             normalText = stream.intermediateDecode()

#             offset = endOffset

#         return Transcription.create(normalText, Token.fromTokenMetadata(text.transcripts[0].tokens))

#     def transcribe(
#         self,
#         dataset: CustomDataset,
#         batchSize: int,
#         ignoreCache: bool = False
#     ) -> TranscriptionList:

#         """
#             Transcribes audio dataset into text and separates
#             text into tokens.
#             If text sample is contained inside the dataset
#             it is also processed and tokenized

#             Parameters
#             ----------
#             dataset : CustomDataset
#                 dataset used for transcription
#             batchSize : int
#                 size of the chunk extracted for transcription
#             ignoreCache : bool
#                 if True cached dataset is ignored

#             Returns
#             -------
#             TranscriptionList -> list of tuples, each tuple contains sample and its transcription
#         """

#         # TODO: Enable once backend is working
#         # if not ignoreCache and _hasCachedTranscription(dataset, self.parameters):
#         #     logging.info(f">> [Coretex] Using cached transcription for: {dataset.name}")
#         #     return _fetchCachedTranscription(dataset, self.parameters)

#         result: TranscriptionList = []

#         for sample in dataset.samples:
#             result.append((sample, self.__transcribeSingle(sample, batchSize)))

#         # TODO: Enable once backend is working
#         # if _cacheTranscription(dataset, result, self.parameters):
#         #     logging.info(f">> [Coretex] Cached transcription for: {dataset.name}")
#         # else:
#         #     logging.info(f">> [Coretex] Failed to cache transcription for: {dataset.name}")

#         return result



from typing import Final, Any, Optional
from pathlib import Path

import json

from PIL import Image, ImageOps

import numpy as np

from .image_format import ImageFormat
#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from ...annotation import CoretexImageAnnotation, ImageDatasetClasses


def _findImage(path: Path) -> Path:
    for format in ImageFormat:
        imagePaths = list(path.glob(f"*.{format.extension}"))
        imagePaths = [path for path in imagePaths if not "thumb" in str(path)]

        if len(imagePaths) > 0:
            return Path(imagePaths[0])

    raise RuntimeError


def _readImageData(path: Path) -> np.ndarray:
    image = ImageOps.exif_transpose(Image.open(path))
    if image.mode != "RGB":
        image = image.convert("RGB")

    imageData = np.frombuffer(image.tobytes(), dtype = np.uint8)
    imageData = imageData.reshape((image.size[1], image.size[0], 3))

    return imageData


def _readAnnotationData(path: Path) -> CoretexImageAnnotation:
    with open(path, "r") as annotationsFile:
        return CoretexImageAnnotation.decode(
            json.load(annotationsFile)
        )


class AnnotatedImageSampleData:

    """
        Contains image data as well as its annotation\n
        Annotation is expected to be in Coretex.ai format
    """

    def __init__(self, path: Path) -> None:
        self.image: Final = _readImageData(_findImage(path))
        self.annotation: Optional[CoretexImageAnnotation] = None

        annotationPath = path / "annotations.json"
        if annotationPath.exists():
            self.annotation = _readAnnotationData(annotationPath)

    def extractSegmentationMask(self, classes: ImageDatasetClasses) -> np.ndarray:
        """
            Generates segmentation mask for the provided classes

            Parameters
            ----------
            classes : ImageDatasetClasses
                list of dataset classes

            Returns
            -------
            np.ndarray -> segmentation mask represented as np.ndarray
        """

        if self.annotation is None:
            raise ValueError

        return self.annotation.extractSegmentationMask(classes)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Union, Type, TypeVar, Optional, Iterator, Any
from http import HTTPStatus

from requests import Response
from requests.structures import CaseInsensitiveDict


JsonType = TypeVar("JsonType", bound = Union[list, dict])


class NetworkResponse:

    """
        Represents Coretex backend response to network request

        Properties
        ----------
        response : Response
            python.requests HTTP reponse
        endpoint : str
            endpoint to which the request was sent
    """

    def __init__(self, response: Response, endpoint: str):
        self._raw = response
        self.endpoint = endpoint

    @property
    def statusCode(self) -> int:
        """
            Status code of the HTTP response
        """

        return self._raw.status_code

    @property
    def headers(self) -> CaseInsensitiveDict:
        """
            HTTP Response headers
        """

        return self._raw.headers

    def hasFailed(self) -> bool:
        """
            Checks if request has failed

            Returns
            -------
            bool -> True if request has failed, False if request has not failed
        """

        return not self._raw.ok

    def isUnauthorized(self) -> bool:
        """
            Checks if request was unauthorized

            Returns
            -------
            bool -> True if status code is 401 and request has failed, False if not
        """

        return self.statusCode == HTTPStatus.UNAUTHORIZED and self.hasFailed()

    def getJson(self, type_: Type[JsonType]) -> JsonType:
        """
            Converts HTTP response body to json

            Parameters
            ----------
                type_: Type[JsonType]
                    list or dict types to which the json should be cast

            Returns
            -------
            JsonType -> Either a list or a dict object depending on type_ parameter

            Raises
            ------
            ValueError -> If "Content-Type" header was not "application/json"
            TypeError -> If it was not possible to convert body to type of passed "type_" parameter
        """

        if not "application/json" in self.headers.get("Content-Type", ""):
            raise ValueError(f">> [Coretex] Trying to convert request response to json but response \"Content-Type\" was \"{self.headers.get('Content-Type')}\"")

        value = self._raw.json()
        if not isinstance(value, type_):
            raise TypeError(f">> [Coretex] Expected json response to be of type \"{type_.__name__}\", received \"{type(value).__name__}\"")

        return value

    def getContent(self) -> bytes:
        """
            Returns
            -------
            bytes -> body of the request as bytes
        """

        return self._raw.content

    def stream(self, chunkSize: Optional[int] = 1, decodeUnicode: bool = False) -> Iterator[Any]:
        """
            Downloads HTTP response in chunks and returns them as they are being downloaded

            Parameters
            ----------
            chunkSize : Optional[int]
                A value of None will function differently depending on the value of stream.
                stream = True will read data as it arrives in whatever size the chunks are
                received. If stream = False, data is returned as a single chunk.
            decodeUnicode : bool
                If decode_unicode is True, content will be decoded using the best
                available encoding based on the response

            Returns
            -------
            Iterator[Any] -> HTTP response as chunks
        """

        return self._raw.iter_content(chunkSize, decodeUnicode)


class NetworkRequestError(Exception):

    """
        Exception which is raised when an request fails.
        Request is marked as failed when the http code is: >= 400
    """

    def __init__(self, response: NetworkResponse, message: str) -> None:
        if not response.hasFailed():
            raise ValueError(">> [Coretex] Invalid request response")

        try:
            # This will raise ValueError if response is not of type application/json
            # which is the case for response (mostly errors) returned by nginx which are html
            responseJson = response.getJson(dict)

            if "message" in responseJson:
                responseMessage = responseJson["message"]
            else:
                responseMessage = response._raw.content.decode()
        except ValueError:
            responseMessage = response._raw.content.decode()

        super().__init__(f">> [Coretex] {message}. Reason: {responseMessage}")

        self.response = response



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Optional, List, Set, Dict
from pathlib import Path

import glob
import os
import json
import logging

import numpy as np

from ..base_converter import BaseConverter
from ...annotation import CoretexImageAnnotation, CoretexSegmentationInstance, BBox


class CityScapeConverter(BaseConverter):

    def __init__(self, datasetName: str, projectId: int, datasetPath: str) -> None:
        super().__init__(datasetName, projectId, datasetPath)

        self.__baseImagePath = os.path.join(datasetPath, "leftImg8bit_trainvaltest", "leftImg8bit")
        self.__baseAnnotationsPaths = [
            os.path.join(datasetPath, "gtFine_trainvaltest", "gtFine", "train"),
            os.path.join(datasetPath, "gtFine_trainvaltest", "gtFine", "val")
        ]

        self.__imagePaths: List[str] = []
        self.__imagePaths.extend(glob.glob(f"{self.__baseImagePath}/train/*/*.png"))
        self.__imagePaths.extend(glob.glob(f"{self.__baseImagePath}/val/*/*.png"))

    def __annotationPathFor(self, imagePath: str) -> str:
        # Extract last 2 components of imagePath
        annotationName = os.path.sep.join(Path(imagePath).parts[-2:])

        # Replace image specific name with annotation name
        annotationName = annotationName.replace("leftImg8bit.png", "gtFine_polygons.json")

        for annotationsPath in self.__baseAnnotationsPaths:
            annotationPath = os.path.join(annotationsPath, annotationName)

            if os.path.exists(annotationPath):
                return annotationPath

        raise RuntimeError

    def _dataSource(self) -> List[str]:
        return self.__imagePaths

    def _extractLabels(self) -> Set[str]:
        labels: Set[str] = set()

        for imagePath in self.__imagePaths:
            annotationPath = self.__annotationPathFor(imagePath)

            with open(annotationPath, mode="r") as annotationFile:
                annotationData: Dict[str, Any] = json.load(annotationFile)

                for obj in annotationData["objects"]:
                    labels.add(obj["label"])

        return labels

    def __extractInstance(self, obj: Dict[str, Any]) -> Optional[CoretexSegmentationInstance]:
        label = obj["label"]

        coretexClass = self._dataset.classByName(label)
        if coretexClass is None:
            logging.getLogger("coretexpylib").info(f">> [Coretex] Class: ({label}) is not a part of dataset")
            return None

        polygon = np.array(obj["polygon"]).flatten().tolist()

        return CoretexSegmentationInstance.create(
            coretexClass.classIds[0],
            BBox.fromPoly(polygon),
            [polygon]
        )

    def __extractImageAnnotation(self, imagePath: str, annotationData: Dict[str, Any]) -> None:
        imageName = Path(imagePath).stem
        width = annotationData["imgWidth"]
        height = annotationData["imgHeight"]

        coretexAnnotation = CoretexImageAnnotation.create(imageName, width, height, [])

        for obj in annotationData["objects"]:
            instance = self.__extractInstance(obj)
            if instance is None:
                continue

            coretexAnnotation.instances.append(instance)

        self._saveImageAnnotationPair(imagePath, coretexAnnotation)

    def _extractSingleAnnotation(self, imagePath: str) -> None:
        annotationPath = self.__annotationPathFor(imagePath)

        with open(annotationPath, mode="r") as annotationFile:
            annotationData: Dict[str, Any] = json.load(annotationFile)
            self.__extractImageAnnotation(imagePath, annotationData)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Final, Optional, List, Dict, Set

import glob
import os
import json
import logging

from ..base_converter import BaseConverter
from ...annotation import CoretexImageAnnotation, CoretexSegmentationInstance, BBox


class _CocoImageAnnotationData:

    def __init__(self, data: Dict[str, Any], imageInfo: Dict[str, Any]) -> None:
        self.data = data
        self.imageInfo = imageInfo


class COCOConverter(BaseConverter):

    def __init__(self, datasetName: str, projectId: int, datasetPath: str) -> None:
        super().__init__(datasetName, projectId, datasetPath)

        self.__imagesPath: Final = os.path.join(datasetPath, "images")

        annotationsPath = os.path.join(datasetPath, "annotations")
        self.__fileNames: Final = glob.glob(os.path.join(annotationsPath, "*.json"))

    def _dataSource(self) -> List[_CocoImageAnnotationData]:
        fullAnnotationData: List[_CocoImageAnnotationData] = []

        for fileName in self.__fileNames:
            with open(fileName) as jsonFile:
                data = json.load(jsonFile)

                fullAnnotationData.extend([
                    _CocoImageAnnotationData(data, imageInfo)
                    for imageInfo in data["images"]
                ])

        return fullAnnotationData

    def _extractLabels(self) -> Set[str]:
        labels: Set[str] = set()

        for fileName in self.__fileNames:
            with open(fileName) as jsonFile:
                data = json.load(jsonFile)

                for category in data["categories"]:
                    labels.add(category["name"])

        return labels

    def __extractInstance(
        self,
        categories: List[Dict[str, Any]],
        annotation: Dict[str, Any]
    ) -> Optional[CoretexSegmentationInstance]:

        label: Optional[str] = None

        for category in categories:
            if category["id"] == annotation["category_id"]:
                label = category["name"]

        if label is None:
            logging.getLogger("coretexpylib").info(f">> [Coretex] Invalid class: {label}")
            return None

        coretexClass = self._dataset.classByName(label)
        if coretexClass is None:
            logging.getLogger("coretexpylib").info(f">> [Coretex] Class: ({label}) is not a part of dataset")
            return None

        bbox = BBox(*(annotation["bbox"]))

        if "segmentation" in annotation:
            segmentation = annotation["segmentation"]
        else:
            segmentation = [
                bbox.polygon
            ]

        return CoretexSegmentationInstance.create(
            coretexClass.classIds[0],
            bbox,
            segmentation
        )

    def _extractSingleAnnotation(self, annotationData: _CocoImageAnnotationData) -> None:
        imageName = annotationData.imageInfo["file_name"]
        width = annotationData.imageInfo["width"]
        height = annotationData.imageInfo["height"]

        imagePath = os.path.join(self.__imagesPath, imageName)
        if not os.path.exists(imagePath):
            return

        coretexAnnotation = CoretexImageAnnotation.create(imageName, width, height, [])

        for annotation in annotationData.data["annotations"]:
            if annotation["image_id"] != annotationData.imageInfo["id"]:
                continue

            instance = self.__extractInstance(annotationData.data["categories"], annotation)
            if instance is None:
                continue

            coretexAnnotation.instances.append(instance)

        self._saveImageAnnotationPair(imagePath, coretexAnnotation)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Union
from pathlib import Path

from .local_sequence_sample import LocalSequenceSample
from ..network_sample import NetworkSample


class SequenceSample(NetworkSample, LocalSequenceSample):

    """
        Represents the local custom Sample class
        which is used for working with Other Task locally
    """

    def __init__(self) -> None:
        NetworkSample.__init__(self)

    @classmethod
    def isValidSequenceFile(cls, path: Union[Path, str]) -> bool:
        """
            Checks whether the file is a valid sequence file or not.
            File is a valid sequence file if it ends with any of these extensions:
                - .fasta
                - .fastq
                - .fa
                - .fq
        """

        if not isinstance(path, Path):
            path = Path(path)

        supportedExtensions = cls.supportedExtensions()
        supportedExtensions.extend([f"{extension}.gz" for extension in cls.supportedExtensions()])

        return any(path.name.endswith(extension) for extension in supportedExtensions)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Dict, Tuple, Optional
from enum import Enum
from pathlib import Path
from base64 import b64encode

import logging

import click

from . import config_defaults
from .utils import isGPUAvailable
from .ui import clickPrompt, arrowPrompt, highlightEcho, errorEcho, progressEcho, successEcho, stdEcho
from .node_mode import NodeMode
from ...cryptography import rsa
from ...networking import networkManager, NetworkRequestError
from ...configuration import loadConfig, saveConfig, isNodeConfigured, getInitScript
from ...utils import CommandException, docker
from ...entities.model import Model


class NodeException(Exception):
    pass


class ImageType(Enum):

    official = "official"
    custom = "custom"


def pull(image: str) -> None:
    try:
        progressEcho(f"Fetching image {image}...")
        docker.imagePull(image)

        successEcho(f"Image {image} successfully fetched.")
    except BaseException as ex:
        logging.getLogger("cli").debug(ex, exc_info = ex)
        raise NodeException("Failed to fetch latest node version.")


def isRunning() -> bool:
    return docker.containerRunning(config_defaults.DOCKER_CONTAINER_NAME)


def exists() -> bool:
    return docker.containerExists(config_defaults.DOCKER_CONTAINER_NAME)


def start(dockerImage: str, config: Dict[str, Any]) -> None:
    try:
        progressEcho("Starting Coretex Node...")
        docker.createNetwork(config_defaults.DOCKER_CONTAINER_NETWORK)

        environ = {
            "CTX_API_URL": config["serverUrl"],
            "CTX_STORAGE_PATH": "/root/.coretex",
            "CTX_NODE_ACCESS_TOKEN": config["nodeAccessToken"],
            "CTX_NODE_MODE": config["nodeMode"]
        }

        modelId = config.get("modelId")
        if isinstance(modelId, int):
            environ["CTX_MODEL_ID"] = modelId

        nodeSecret = config.get("nodeSecret", config_defaults.DEFAULT_NODE_SECRET)
        if isinstance(nodeSecret, str) and nodeSecret != config_defaults.DEFAULT_NODE_SECRET:
            environ["CTX_NODE_SECRET"] = nodeSecret

        volumes = [
            (config["storagePath"], "/root/.coretex")
        ]

        if config.get("allowDocker", False):
            volumes.append(("/var/run/docker.sock", "/var/run/docker.sock"))

        initScript = getInitScript(config)
        if initScript is not None:
            volumes.append((str(initScript), "/script/init.sh"))

        docker.start(
            config_defaults.DOCKER_CONTAINER_NAME,
            dockerImage,
            config["allowGpu"],
            config["nodeRam"],
            config["nodeSwap"],
            config["nodeSharedMemory"],
            config["cpuCount"],
            environ,
            volumes
        )

        successEcho("Successfully started Coretex Node.")
    except BaseException as ex:
        logging.getLogger("cli").debug(ex, exc_info = ex)
        raise NodeException("Failed to start Coretex Node.")


def clean() -> None:
    try:
        docker.removeContainer(config_defaults.DOCKER_CONTAINER_NAME)
        docker.removeNetwork(config_defaults.DOCKER_CONTAINER_NETWORK)
    except BaseException as ex:
        logging.getLogger("cli").debug(ex, exc_info = ex)
        raise NodeException("Failed to clean inactive Coretex Node.")


def stop() -> None:
    try:
        progressEcho("Stopping Coretex Node...")
        docker.stopContainer(config_defaults.DOCKER_CONTAINER_NAME)
        clean()
        successEcho("Successfully stopped Coretex Node....")
    except BaseException as ex:
        logging.getLogger("cli").debug(ex, exc_info = ex)
        raise NodeException("Failed to stop Coretex Node.")


def getRepoFromImageUrl(image: str) -> str:
    imageName = image.split("/")[-1]
    if not ":" in imageName:
        return image

    tagIndex = image.rfind(":")
    if tagIndex != -1:
        return image[:tagIndex]
    else:
        return image


def getTagFromImageUrl(image: str) -> str:
    imageName = image.split("/")[-1]
    if not ":" in imageName:
        return "latest"

    tagIndex = image.rfind(":")
    if tagIndex != -1:
        return image[tagIndex + 1:]
    else:
        return "latest"


def shouldUpdate(image: str) -> bool:
    repository = getRepoFromImageUrl(image)
    try:
        imageJson = docker.imageInspect(image)
    except CommandException:
        # imageInspect() will raise an error if image doesn't exist locally
        return True

    try:
        manifestJson = docker.manifestInspect(image)
    except CommandException:
        return False

    for digest in imageJson["RepoDigests"]:
        if repository in digest and manifestJson["Descriptor"]["digest"] in digest:
            return False

    return True


def registerNode(name: str, publicKey: Optional[bytes] = None, nearWalletId: Optional[str] = None) -> str:
    params = {
        "machine_name": name,
    }

    if publicKey is not None:
        params["public_key"] = b64encode(publicKey).decode("utf-8")

    if nearWalletId is not None:
        params["near_wallet_id"] = nearWalletId

    response = networkManager.post("service", params)

    if response.hasFailed():
        raise NetworkRequestError(response, "Failed to configure node. Please try again...")

    accessToken = response.getJson(dict).get("access_token")

    if not isinstance(accessToken, str):
        raise TypeError("Something went wrong. Please try again...")

    return accessToken


def selectImageType() -> ImageType:
    availableImages = {
        "Official Coretex image": ImageType.official,
        "Custom image": ImageType.custom,
    }

    choices = list(availableImages.keys())
    selectedImage = arrowPrompt(choices, "Please select image that you want to use (use arrow keys to select an option):")

    return availableImages[selectedImage]


def selectModelId(storagePath: str, retryCount: int = 0) -> int:
    if retryCount >= 3:
        raise RuntimeError("Failed to fetch Coretex Model. Terminating...")

    modelId: int = clickPrompt("Specify Coretex Model ID that you want to use:", type = int)

    try:
        model = Model.fetchById(modelId)
    except:
        errorEcho(f"Failed to fetch model with id {modelId}.")
        return selectModelId(storagePath, retryCount + 1)

    modelDir = Path(storagePath) / "models"
    modelDir.mkdir(parents = True, exist_ok = True)
    model.download(modelDir / str(model.id))

    return modelId


def selectNodeMode(storagePath: str) -> Tuple[int, Optional[int]]:
    availableNodeModes = {
        "Run workflows (worker)": NodeMode.execution,
        "Serve a single endpoint (dedicated inference)": NodeMode.functionExclusive,
        "Serve multiple endpoints (shared inference)": NodeMode.functionShared
    }
    choices = list(availableNodeModes.keys())

    selectedMode = arrowPrompt(choices, "Please select Coretex Node mode (use arrow keys to select an option):")

    if availableNodeModes[selectedMode] == NodeMode.functionExclusive:
        modelId = selectModelId(storagePath)
        return availableNodeModes[selectedMode], modelId

    return availableNodeModes[selectedMode], None


def promptCpu(config: Dict[str, Any], cpuLimit: int) -> int:
    cpuCount: int = clickPrompt(f"Enter the number of CPUs the container will use (Maximum: {cpuLimit}) (press enter to use default)", cpuLimit, type = int)

    if cpuCount == 0:
        errorEcho(f"ERROR: Number of CPU's the container will use must be higher than 0")
        return promptCpu(config, cpuLimit)

    if cpuCount > cpuLimit:
        errorEcho(f"ERROR: CPU limit in Docker Desktop ({cpuLimit}) is lower than the specified value ({cpuCount})")
        return promptCpu(config, cpuLimit)

    return cpuCount


def promptRam(config: Dict[str, Any], ramLimit: int) -> int:
    nodeRam: int = clickPrompt(f"Node RAM memory limit in GB (Minimum: {config_defaults.MINIMUM_RAM}GB, Maximum: {ramLimit}GB) (press enter to use default)", ramLimit, type = int)

    if nodeRam > ramLimit:
        errorEcho(f"ERROR: RAM limit in Docker Desktop ({ramLimit}GB) is lower than the configured value ({config['nodeRam']}GB). Please adjust resource limitations in Docker Desktop settings.")
        return promptRam(config, ramLimit)

    if nodeRam < config_defaults.MINIMUM_RAM:
        errorEcho(f"ERROR: Configured RAM ({nodeRam}GB) is lower than the minimum Node RAM requirement ({config_defaults.MINIMUM_RAM}GB).")
        return promptRam(config, ramLimit)

    return nodeRam


def _configureInitScript() -> str:
    initScript = clickPrompt("Enter a path to sh script which will be executed before Node starts", config_defaults.DEFAULT_INIT_SCRIPT, type = str)

    if initScript == config_defaults.DEFAULT_INIT_SCRIPT:
        return config_defaults.DEFAULT_INIT_SCRIPT

    path = Path(initScript).expanduser().absolute()

    if path.is_dir():
        errorEcho("Provided path is pointing to a directory, file expected!")
        return _configureInitScript()

    if not path.exists():
        errorEcho("Provided file does not exist!")
        return _configureInitScript()

    return str(path)


def checkResourceLimitations() -> None:
    _, ramLimit = docker.getResourceLimits()

    if ramLimit < config_defaults.MINIMUM_RAM:
        raise RuntimeError(f"Minimum Node RAM requirement ({config_defaults.MINIMUM_RAM}GB) is higher than your current Docker desktop RAM limit ({ramLimit}GB). Please adjust resource limitations in Docker Desktop settings to match Node requirements.")


def isConfigurationValid(config: Dict[str, Any]) -> bool:
    isValid = True
    cpuLimit, ramLimit = docker.getResourceLimits()

    if not isinstance(config["nodeRam"], int):
        errorEcho(f"Invalid config \"nodeRam\" field type \"{type(config['nodeRam'])}\". Expected: \"int\"")
        isValid = False

    if not isinstance(config["cpuCount"], int):
        errorEcho(f"Invalid config \"cpuCount\" field type \"{type(config['cpuCount'])}\". Expected: \"int\"")
        isValid = False

    if config["cpuCount"] > cpuLimit:
        errorEcho(f"Configuration not valid. CPU limit in Docker Desktop ({cpuLimit}) is lower than the configured value ({config['cpuCount']})")
        isValid = False

    if ramLimit < config_defaults.MINIMUM_RAM:
        errorEcho(f"Minimum Node RAM requirement ({config_defaults.MINIMUM_RAM}GB) is higher than your current Docker desktop RAM limit ({ramLimit}GB). Please adjust resource limitations in Docker Desktop settings to match Node requirements.")
        isValid = False

    if config["nodeRam"] > ramLimit:
        errorEcho(f"Configuration not valid. RAM limit in Docker Desktop ({ramLimit}GB) is lower than the configured value ({config['nodeRam']}GB)")
        isValid = False


    if config["nodeRam"] < config_defaults.MINIMUM_RAM:
        errorEcho(f"Configuration not valid. Minimum Node RAM requirement ({config_defaults.MINIMUM_RAM}GB) is higher than the configured value ({config['nodeRam']}GB)")
        isValid = False

    return isValid


def configureNode(config: Dict[str, Any], verbose: bool) -> None:
    highlightEcho("[Node Configuration]")

    cpuLimit, ramLimit = docker.getResourceLimits()

    config["nodeName"] = clickPrompt("Node name", type = str)

    imageType = selectImageType()
    if imageType == ImageType.custom:
        config["image"] = clickPrompt("Specify URL of docker image that you want to use:", type = str)
    else:
        config["image"] = "coretexai/coretex-node"

    if isGPUAvailable():
        config["allowGpu"] = clickPrompt("Do you want to allow the Node to access your GPU? (Y/n)", type = bool, default = True)
    else:
        config["allowGpu"] = False

    if imageType == ImageType.official:
        tag = "gpu" if config["allowGpu"] else "cpu"
        config["image"] += f":latest-{tag}"

    config["storagePath"] = config_defaults.DEFAULT_STORAGE_PATH
    config["nodeRam"] = int(min(max(config_defaults.MINIMUM_RAM, ramLimit), config_defaults.DEFAULT_RAM))
    config["nodeSwap"] = config_defaults.DEFAULT_SWAP_MEMORY
    config["nodeSharedMemory"] = config_defaults.DEFAULT_SHARED_MEMORY
    config["cpuCount"] = int(min(cpuLimit, config_defaults.DEFAULT_CPU_COUNT))
    config["nodeMode"] = config_defaults.DEFAULT_NODE_MODE
    config["allowDocker"] = config_defaults.DEFAULT_ALLOW_DOCKER
    config["nodeSecret"] = config_defaults.DEFAULT_NODE_SECRET
    config["initScript"] = config_defaults.DEFAULT_INIT_SCRIPT

    publicKey: Optional[bytes] = None
    nearWalletId: Optional[str] = None

    if verbose:
        config["storagePath"] = clickPrompt("Storage path (press enter to use default)", config_defaults.DEFAULT_STORAGE_PATH, type = str)

        config["cpuCount"] = promptCpu(config, cpuLimit)

        config["nodeRam"] = promptRam(config, ramLimit)

        config["nodeSwap"] = clickPrompt("Node swap memory limit in GB, make sure it is larger than mem limit (press enter to use default)", config_defaults.DEFAULT_SWAP_MEMORY, type = int)
        config["nodeSharedMemory"] = clickPrompt("Node POSIX shared memory limit in GB (press enter to use default)", config_defaults.DEFAULT_SHARED_MEMORY, type = int)
        config["allowDocker"] = clickPrompt("Allow Node to access system docker? This is a security risk! (Y/n)", config_defaults.DEFAULT_ALLOW_DOCKER, type = bool)
        config["initScript"] = _configureInitScript()

        nodeMode, modelId = selectNodeMode(config["storagePath"])
        config["nodeMode"] = nodeMode
        if modelId is not None:
            config["modelId"] = modelId

        nodeSecret: str = clickPrompt("Enter a secret which will be used to generate RSA key-pair for Node", config_defaults.DEFAULT_NODE_SECRET, type = str, hide_input = True)
        config["nodeSecret"] = nodeSecret

        if nodeSecret != config_defaults.DEFAULT_NODE_SECRET:
            progressEcho("Generating RSA key-pair (2048 bits long) using provided node secret...")
            rsaKey = rsa.generateKey(2048, nodeSecret.encode("utf-8"))
            publicKey = rsa.getPublicKeyBytes(rsaKey.public_key())

        if nodeMode != NodeMode.execution:
            nearWalletId = clickPrompt(
                "Enter a NEAR wallet id to which the funds will be transfered when executing endpoints",
                config_defaults.DEFAULT_NEAR_WALLET_ID,
                type = str
            )

            if nearWalletId != config_defaults.DEFAULT_NEAR_WALLET_ID:
                config["nearWalletId"] = nearWalletId
            else:
                config["nearWalletId"] = None
                nearWalletId = None
    else:
        stdEcho("To configure node manually run coretex node config with --verbose flag.")

    config["nodeAccessToken"] = registerNode(config["nodeName"], publicKey, nearWalletId)


def initializeNodeConfiguration() -> None:
    config = loadConfig()
    isConfigured = isNodeConfigured(config)

    if isConfigured:
        if isConfigurationValid(config):
            return

        errorEcho("Invalid node configuration found.")
        if not click.confirm("Would you like to update the configuration?", default = True):
            raise RuntimeError("Invalid configuration. Please use \"coretex node config\" to update a Node configuration.")

    if not isConfigured:
        errorEcho("Node configuration not found.")

    if isRunning():
        if not click.confirm("Node is already running. Do you wish to stop the Node?", default = True):
            errorEcho("If you wish to reconfigure your node, use \"coretex node stop\" command first.")
            return

        stop()

    configureNode(config, verbose = False)
    saveConfig(config)



from typing import Optional

import click

from ..modules import project_utils, user, utils, ui
from ...entities import Model
from ...configuration import loadConfig


@click.command()
@click.argument("path", type = click.Path(exists = True, file_okay = False, dir_okay = True))
@click.option("-n", "--name", type = str, required = True)
@click.option("-p", "--project", type = str, required = False, default = None)
@click.option("-a", "--accuracy", type = click.FloatRange(0, 1), required = False, default = 1)
def create(name: str, path: str, project: Optional[str], accuracy: float) -> None:
    config = loadConfig()

    # If project was provided used that, otherwise get the one from config
    # If project that was provided does not exist prompt user to create a new
    # one with that name
    ctxProject = project_utils.getProject(project, config)
    if ctxProject is None:
        return

    ui.progressEcho("Creating the model...")

    model = Model.createProjectModel(name, ctxProject.id, accuracy)
    model.upload(path)

    ui.successEcho(f"Model \"{model.name}\" created successfully")
    ui.stdEcho(f"A new model has been created. You can open it by clicking on this URL {ui.outputUrl(model.entityUrl())}.")


@click.group()
@utils.onBeforeCommandExecute(user.initializeUserSession)
def model() -> None:
    pass


model.add_command(create)



from pathlib import Path

import sys
import logging
import runpy

from coretex import _task
from coretex.networking import RequestFailedError


if __name__ == "__main__":
    taskRun, callback = _task.processRemote(sys.argv)

    try:
        _task._prepareForExecution(taskRun)
        _task.current_task_run.setCurrentTaskRun(taskRun)

        callback.onStart()

        logging.getLogger("coretexpylib").info(">> [Coretex] TaskRun execution started")
        logging.getLogger("coretexpylib").info(f"\tPython: {sys.executable}")

        entryPointDir = str(Path(taskRun.entryPoint).parent)
        if entryPointDir not in sys.path:
            sys.path.append(entryPointDir)

        _coretexPath = str(Path(__file__).resolve().parent)
        if _coretexPath in sys.path:
            sys.path.remove(_coretexPath)

        # Run the entry point script
        runpy.run_path(taskRun.entryPoint, {}, "__main__")

        callback.onSuccess()
    except RequestFailedError:
        callback.onNetworkConnectionLost()
    except KeyboardInterrupt:
        callback.onKeyboardInterrupt()
    except BaseException as ex:
        callback.onException(ex)

        # sys.exit is ok here, finally block is guaranteed to execute
        # due to how sys.exit is implemented (it internally raises SystemExit exception)
        sys.exit(1)
    finally:
        callback.onCleanUp()



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, List, Optional, TypeVar, Type

import os

from ..parameter_type import ParameterType
from ..base_parameter import BaseParameter
from ....project import ProjectType
from ....secret import Secret


T = TypeVar("T", bound = Secret)


class SecretParameter(BaseParameter[str]):

    def __init__(
        self,
        secretType: Type[T],
        name: str,
        description: str,
        value: Optional[str],
        dataType: ParameterType,
        required: bool,
        type: int
    ) -> None:

        self._secretType = secretType

        super().__init__(name, description, value, dataType, required, type)

    @property
    def types(self) -> List[type]:
        return [str]

    def parseValue(self, type_: ProjectType) -> Optional[Any]:
        if self.value is None:
            return None

        if "CTX_NODE_ACCESS_TOKEN" in os.environ:
            return self._secretType.fetchNodeSecret(self.value, os.environ["CTX_NODE_ACCESS_TOKEN"])

        return self._secretType.fetchByName(self.value)



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional

from .network_response import NetworkResponse
from .network_manager_base import NetworkManagerBase
from .user_data import UserData


class NetworkManager(NetworkManagerBase):

    """
        Subclass of NetworkManagerBase intended to be used
        by a normal user. Contains functionality related to user
        and authenticating as a normal user.
    """

    def __init__(self) -> None:
        super().__init__()

        self.__userData = UserData()

    @property
    def _apiToken(self) -> Optional[str]:
        return self.__userData.apiToken

    @_apiToken.setter
    def _apiToken(self, value: Optional[str]) -> None:
        self.__userData.apiToken = value

    @property
    def _refreshToken(self) -> Optional[str]:
        return self.__userData.refreshToken

    @_refreshToken.setter
    def _refreshToken(self, value: Optional[str]) -> None:
        self.__userData.refreshToken = value

    @property
    def hasStoredCredentials(self) -> bool:
        return self.__userData.hasStoredCredentials

    def authenticate(self, username: str, password: str, storeCredentials: bool = True) -> NetworkResponse:
        """
            Authenticates user with provided credentials

            Parameters
            ----------
            username : str
                Coretex.ai username
            password : str
                Coretex.ai password
            storeCredentials : bool
                If true credentials will be stored in User object for reuse

            Returns
            -------
            NetworkResponse -> NetworkResponse object containing the full response info

            Example
            -------
            >>> from coretex.networking import networkManager
            \b
            >>> response = networkManager.authenticate(username = "dummy@coretex.ai", password = "123456")
            >>> if response.hasFailed():
                    print("Failed to authenticate")
        """

        if storeCredentials:
            self.__userData.username = username
            self.__userData.password = password

        # authenticate using credentials stored in requests.Session.auth
        return super().authenticate(username, password, storeCredentials)

    def authenticateWithStoredCredentials(self) -> NetworkResponse:
        """
            Authenticates user with credentials stored inside User object

            Returns
            -------
            NetworkResponse -> NetworkResponse object containing the full response info

            Raises
            ------
            ValueError -> if credentials are not found
        """

        if self.__userData.username is None or self.__userData.password is None:
            raise ValueError(">> [Coretex] Credentials not stored")

        return self.authenticate(self.__userData.username, self.__userData.password)


networkManager: NetworkManagerBase = NetworkManager()



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Optional
from pathlib import Path

import sys
import logging

from .formatter import CTXFormatter
from .severity import LogSeverity
from ..utils import createFileHandler


def createFormatter(
    includeTime: bool = True,
    includeLevel: bool = True,
    includeColor: bool = True,
    jsonOutput: bool = True
) -> logging.Formatter:

    fmt: List[str] = []

    if includeTime:
        fmt.append("%(asctime)s")

    if includeLevel:
        fmt.append("%(levelname)s")

    if len(fmt) != 0:
        fmt[-1] = fmt[-1] + ":"

    fmt.append("%(message)s")

    return CTXFormatter(
        fmt = " ".join(fmt),
        datefmt= "%Y-%m-%d %H:%M:%S",
        style = "%",
        color = includeColor,
        jsonOutput = jsonOutput
    )


def initializeLogger(
    severity: LogSeverity,
    logPath: Path,
    streamHandler: Optional[logging.StreamHandler] = None,
    jsonOutput: bool = False
) -> None:

    """
        Initializes python logging module

        Parameters
        ----------
        severity : LogSeverity
            Severity level of the logger
        logPath : Path
            File path where logs will be stored
        streamHandler: Optional[logging.StreamHandler] : bool
            Stream handler which will be used for logging
            If None default one will be used
    """

    if streamHandler is None:
        streamHandler = logging.StreamHandler(sys.stdout)
        streamHandler.setLevel(severity.getLevel())

        streamHandler.setFormatter(createFormatter(
            includeTime = False,
            jsonOutput = jsonOutput
        ))

    fileHandler = createFileHandler(logPath)
    fileHandler.setLevel(logging.DEBUG)
    fileHandler.setFormatter(createFormatter(includeColor = False, jsonOutput = False))

    logging.basicConfig(
        level = logging.NOTSET,
        force = True,
        handlers = [
            streamHandler,
            fileHandler
        ]
    )



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, Callable, Optional, List
from concurrent.futures import ThreadPoolExecutor, Future

import logging
import os


MAX_WORKER_COUNT = 8


class MultithreadedDataProcessor:

    """
        Used for splitting data processing on multiple threads
        Only useful for I/O bound operations, do not use for
        heavy data processing operations

        data: List[Any]
            list which will be split-processed on multiple threads
        singleElementProcessor: Callable[[Any], None]
            function which will be called for a single element from the provided list
        threadCount: Optional[int]
            number of threads which will be used for processing (maximum is 8), if None
            it will use min(cpuCores, 8)
        title: Optional[str]
            message which is displayed when the processing starts

        Example
        -------
        >>> import os
        >>> from pathlib import Path
        >>> from coretex import CustomItem
        >>> from coretex.networking import networkManager
        >>> from coretex.threading import MultithreadedDataProcessor
        \b
        >>> networkManager.authenticate("dummyUser@coretex.ai", *******) # login on coretex
        >>> path = "" #path to dir which contains samples for upload to dataset
        >>> listOfFiles = []
        \b
        >>> def createItem(path: str) -> CustomItem:
                item = CustomItem.createCustomItem(Path(path).stem, 1023, path)  # first make new dataset
                if item is None:
                    pass
        \b
        >>> for file in os.listdir(path):
                if file.endswith(".zip"):
                    file = os.path.join(path, file)
                    listOfFiles.append(file)
        \b
        >>> processor = MultithreadedDataProcessor(
                listOfFiles,
                createItem,
                threadCount = 8, # set threadCount on desired value default is 8
                title = "createItem"
            )
        >>> processor.process()
    """

    def __init__(
        self,
        data: List[Any],
        singleElementProcessor: Callable[[Any], None],
        workerCount: Optional[int] = None,
        message: Optional[str] = None
    ) -> None:

        if workerCount is None:
            cpuCount = os.cpu_count()

            if cpuCount is None:
                cpuCount = 1

            workerCount = min(MAX_WORKER_COUNT, cpuCount)
        else:
            if workerCount > MAX_WORKER_COUNT:
                logging.warning(f">> [Coretex] \"workerCount\" value: {workerCount} is higher than maximum allowed: {MAX_WORKER_COUNT}. Using {MAX_WORKER_COUNT} workers")

            workerCount = min(MAX_WORKER_COUNT, workerCount)

        self.__data = data
        self.__singleElementProcessor = singleElementProcessor
        self.__workerCount = workerCount
        self.__message = message

    def process(self) -> None:
        """
            Starts the multithreaded processing of the data

            Raises
            ------
            Any unhandled exception which happened during the processing
        """

        if self.__message is not None:
            logging.getLogger("coretexpylib").info(f">> [Coretex] {self.__message}")
            logging.getLogger("coretexpylib").info(f"\tUsing {self.__workerCount} workers")

        futures: List[Future] = []

        with ThreadPoolExecutor(max_workers = self.__workerCount) as pool:
            for element in self.__data:
                future = pool.submit(self.__singleElementProcessor, element)
                futures.append(future)

        for future in futures:
            exception = future.exception()
            if exception is not None:
                raise exception



#     Copyright (C) 2023  Coretex LLC

#     This file is part of Coretex.ai

#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU Affero General Public License as
#     published by the Free Software Foundation, either version 3 of the
#     License, or (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Affero General Public License for more details.

#     You should have received a copy of the GNU Affero General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Optional, Any, Tuple, Dict, List

from .base_parameter import ParameterType


def validateEnumStructure(name: str, value: Optional[Any], required: bool) -> Tuple[bool, Optional[str]]:
    if not isinstance(value, dict):
        return False, None

    # Enum parameter must contain 2 key-value pairs: selected and options
    if len(value) != 2 or "options" not in value or "selected" not in value:
        keys = ", ".join(value.keys())
        return False, f"Enum parameter \"{name}\" must contain only \"selected\" and \"options\" properties, but it contains \"{keys}\""

    options = value.get("options")

    # options must be an object of type list
    if not isinstance(options, list):
        return False, f"Enum parameter \"{name}.options\" has invalid type. Expected \"list[str]\", got \"{type(options).__name__}\""

    # all elements of options list must be strings
    if not all(isinstance(element, str) for element in options):
        elementTypes = ", ".join({type(element).__name__ for element in options})
        return False, f"Elements of enum parameter \"{name}.options\" have invalid type. Expected \"list[str]\" got \"list[{elementTypes}]\""

    # options elements must not be empty strings
    if not all(element != "" for element in options):
        return False, f"Elements of enum parameter \"{name}.options\" must be non-empty strings."

    selected = value.get("selected")
    if selected is None and required:
        return False, f"Enum parameter \"{name}.selected\" has invalid type. Expected \"int\", got \"{type(selected).__name__}\""

    return True, None


def validateRangeStructure(name: str, value: Dict[str, Any], required: bool) -> Tuple[bool, Optional[str]]:
    if len(value) != 3 or "from" not in value or "to" not in value or "step" not in value:
        keys = ", ".join(value.keys())
        return False, f"Range parameter \"{name}\" must contain only \"from\",  \"to\" and \"step\" properties, but it contains \"{keys}\""

    if any(element is None for element in value.values()) and required:
        return False, f"Elements of range parameter \"{name}\" must not be null"

    if not all(type(element) is int for element in value.values()) and required:
        elementTypes = ", ".join({type(element).__name__ for element in value.values()})
        return False, f"Elements of range parameter \"{name}\" have invalid type. Expected \"int\" got \"{elementTypes}\""

    if any(type(element) is float for element in value.values()):
        return False, "Range parameter does not support float values"

    return True, None

def getValueParamType(value: Any, name: str) -> ParameterType:
    if isinstance(value, bool):
        return ParameterType.boolean

    if isinstance(value, int):
        return ParameterType.integer

    if isinstance(value, float):
        return ParameterType.floatingPoint

    if isinstance(value, str):
        return ParameterType.string

    if isinstance(value, list):
        return getListParamType(value, name)

    supportedTypes = [type_.name for type_ in ParameterType]

    raise ValueError(f">> [Coretex] Parameter \"{name}\" has invalid type. Expected \"{supportedTypes}\", got \"{type(value)}\".")

def getListParamType(value: List[Any], name: str) -> ParameterType:
    if all(isinstance(item, int) for item in value):
        return ParameterType.intList

    if all(isinstance(item, float) for item in value):
        return ParameterType.floatList

    if all(isinstance(item, str) for item in value):
        return ParameterType.strList

    typesFound = ", ".join([type(item).__name__ for item in value])
    raise ValueError(f">> [Coretex] Parameter \"{name}\" cannot contain multiple value type: \"{typesFound}\".")
